"""Common Errors for the MapProof Module."""
from typing import Dict, Any
from enum import Enum, auto as enum_auto

from .constants import KEY_SIZE

# Methods are self-documenting here.
# pylint: disable=missing-docstring


class MalformedMapProofError(Exception):
    """Error to be raised if the provided proof is malformed."""

    class ErrorKind(Enum):
        """Kind of the error."""

        EMBEDDED_PATH = enum_auto()
        DUPLICATE_PATH = enum_auto()
        INVALID_ORDERING = enum_auto()
        NON_TERMINAL_NODE = enum_auto()
        MALFORMED_ENTRY = enum_auto()
        INVALID_KEY_SIZE = enum_auto()

    def __init__(self, message: str, error_data: Dict[str, Any]) -> None:
        super().__init__(message)

        self.error_data = error_data

    @classmethod
    def embedded_paths(cls, prefix: Any, path: Any) -> "MalformedMapProofError":
        error_msg = "Embedded path: prefix {}, path {}".format(prefix, path)
        error_data = {"kind": cls.ErrorKind.EMBEDDED_PATH, "prefix": prefix, "path": path}

        return cls(error_msg, error_data)

    @classmethod
    def duplicate_path(cls, path: Any) -> "MalformedMapProofError":
        error_msg = "Duplicate path: path {}".format(path)
        error_data = {"kind": cls.ErrorKind.DUPLICATE_PATH, "path": path}

        return cls(error_msg, error_data)

    @classmethod
    def invalid_ordering(cls, path_a: Any, path_b: Any) -> "MalformedMapProofError":
        error_msg = "Invalid ordering: prev_path {}, path {}".format(path_a, path_b)
        error_data = {"kind": cls.ErrorKind.INVALID_ORDERING, "prev_path": path_a, "path": path_b}

        return cls(error_msg, error_data)

    @classmethod
    def non_terminal_node(cls, node: Any) -> "MalformedMapProofError":
        error_msg = "Non-terminal node: node {}".format(node)
        error_data = {"kind": cls.ErrorKind.NON_TERMINAL_NODE, "node": node}

        return cls(error_msg, error_data)

    @classmethod
    def malformed_entry(cls, entry: Any, additional_info: Any = None) -> "MalformedMapProofError":
        error_msg = "Malformed proof entry: entry {}".format(entry)
        if additional_info:
            error_msg += " [{}]".format(additional_info)
        error_data = {"kind": cls.ErrorKind.MALFORMED_ENTRY, "entry": entry}

        return cls(error_msg, error_data)

    @classmethod
    def invalid_key_size(cls, key: bytes) -> "MalformedMapProofError":
        error_msg = f"Invalid key '{key!r}' for raw MapProof, expected size {KEY_SIZE}, actual {len(key)}"

        error_data = {"kind": cls.ErrorKind.INVALID_KEY_SIZE, "key": key}

        return cls(error_msg, error_data)


class MapProofBuilderError(Exception):
    """Error that can occur in MapProofBuilder."""

    def __init__(self, msg: str, error_data: Any = None) -> None:
        super().__init__(msg)

        self.error_data = error_data



"""MapProofBuilder module."""
from typing import Optional, Dict, Any
from logging import getLogger

from exonum_client_cis.module_manager import ModuleManager
from .map_proof import MapProof
from .errors import MapProofBuilderError
from ..encoder import build_encoder_function

# pylint: disable=C0103
logger = getLogger(__name__)


class MapProofBuilder:
    """
    Builder for MapProof.

    This class is capable of creating MapProof from Dict[Any, Any].
    Since MapProof converts both keys and values to bytes for hashing, this
    class generates converter functions from the relevant Protobuf structures.

    Designed workflow example:

    >>> proof_builder = MapProofBuilder()
    >>> proof_builder.set_key_encoder('BitVec', main_module='types')
    >>> proof_builder.set_value_encoder('Wallet', service_name='cryptocurrency:0.11.0', service_module='service')
    >>> proof = proof_builder.build_proof(map_proof)

    If the proof is built from `RawProofMapIndex`, you can provide bool `raw` variable to the constructor:

    >>> proof_builder = MapProofBuilder(raw=True)

    If your key/value type does not require convertion to bytes through Protobuf, use `MapProof.parse`
    instead and provide your converter functions by yourself.

    If one of the types requires manual convertion, but the other one should use Protobuf, you can
    use `build_encoder_function` function from encoder.py which takes a Protobuf message class as an argument and
    returns a converter function.
    """

    def __init__(self, raw: bool = False) -> None:
        """ Constructor of MapProofBuilder. """
        self._key_encoder: Optional[type] = None
        self._value_encoder: Optional[type] = None
        self._raw = raw

    @staticmethod
    def _get_encoder(
        structure_name: str,
        main_module: Optional[str] = None,
        service_name: Optional[str] = None,
        service_module: Optional[str] = None,
    ) -> type:
        try:
            if main_module:
                module = ModuleManager.import_main_module(main_module)
            elif service_name and service_module:
                module = ModuleManager.import_service_module(service_name, service_module)
            else:
                err = MapProofBuilderError("Module data not provided")
                logger.warning(str(err))
                raise err

            encoder = getattr(module, structure_name)
            logger.debug("Successfully got encoder.")
            return encoder
        except (ModuleNotFoundError, ImportError):
            error_data = {"main_module": main_module, "service_name": service_name, "service_module": service_module}

            err = MapProofBuilderError("Incorrect module data", error_data)
            logger.warning("%s: %s", str(err), error_data)
            raise err
        except AttributeError:
            error_data = {"service_name": structure_name}

            err = MapProofBuilderError("Incorrect structure name", error_data)
            logger.warning("%s: %s", str(err), error_data)
            raise err

    # pylint: disable=too-many-arguments
    def set_key_encoder(
        self,
        structure_name: str,
        main_module: Optional[str] = None,
        service_name: Optional[str] = None,
        service_module: Optional[str] = None,
    ) -> "MapProofBuilder":
        """
        Method to set the key encoder.

        If the Protobuf structure for the key lies in the main module,
        provide the `main_module` argument.
        Otherwise, provide both `service_name` and `service_module`.

        Parameters
        ----------
        structure_name: str
            Name of the Protobuf structure to be used in the converted function.
        main_module: Optional[str]
            Name of the main module.
        service_name: Optional[str]
            Name of the service.
        service_module: Optional[str]
            Name of the service module.

        Raises
        ------
        MapProofBuilderError
            If the provided data is incorrect, this exception rises.
        """
        self._key_encoder = self._get_encoder(
            structure_name, main_module, service_name, service_module
        )
        return self

    # pylint: disable=too-many-arguments
    def set_value_encoder(
        self,
        structure_name: str,
        main_module: Optional[str] = None,
        service_name: Optional[str] = None,
        service_module: Optional[str] = None,
    ) -> "MapProofBuilder":
        """
        Method to set the value encoder.

        If the Protobuf structure for the value lies in the main module,
        provide the `main_module` argument.
        Otherwise, provide both `service_name` and `service_module`.

        Parameters
        ----------
        structure_name: str
            Name of the Protobuf structure to be used in the converted function.
        main_module: Optional[str]
            Name of the main module.
        service_name: Optional[str]
            Name of the service.
        service_module: Optional[str]
            Name of the service module.

        Raises
        ------
        MapProofBuilderError
            If the provided data is incorrect, this exception rises.
        """
        self._value_encoder = self._get_encoder(
            structure_name, main_module, service_name, service_module
        )

        return self

    def build_proof(self, proof: Dict[Any, Any]) -> MapProof:
        """
        Method to build MapProof from Dict[Any, Any].

        Call this method only after `set_key_encoder` and `set_value_encoder`.

        Parameters
        ----------
        proof: Dict[Any, Any]
            Raw ProofMap that will be parsed.

        Raises
        ------
        MapProofBuilderError
            If the method is called without prior `set_key_encoder` and `set_value_encoder` calls,
            this exception will rises.

        MalformedMapProofError
            If provided raw MapProof is malformed and parsing fails, this exception
            rises.
        """
        if not self._key_encoder or not self._value_encoder:
            err = MapProofBuilderError("Encoders are not set.")
            logger.warning(str(err))
            raise err

        key_encoder_func = build_encoder_function(self._key_encoder)
        value_encoder_func = build_encoder_function(self._value_encoder)

        return MapProof.parse(proof, key_encoder_func, value_encoder_func, self._raw)



"""Module with the Bindings to Protoc."""
from typing import List, Optional
from logging import getLogger
import os
import re
import shutil
import subprocess

# pylint: disable=C0103
logger = getLogger(__name__)

PROTOC_ENV_NAME = "PROTOC"
PROTOC_MIN_VERSION = (3, 6, 1)


def _find_protoc() -> Optional[str]:
    if PROTOC_ENV_NAME in os.environ:
        return os.getenv(PROTOC_ENV_NAME)

    return shutil.which("protoc")


def _find_proto_files(path: str) -> List[str]:
    return [file for file in os.listdir(path) if file.endswith(".proto")]


class Protoc:
    """Class that provides an interface to protoc.
    It is not intended to be used in user code, see ProtobufLoader instead."""

    def __init__(self) -> None:
        protoc_path = _find_protoc()
        if protoc_path is None:
            logger.critical("Protobuf compiler not found.")
            raise RuntimeError("protoc was not found, make sure that it is installed")

        self._protoc_path = protoc_path

        self._ensure_protoc_version()

    @staticmethod
    def _modify_file(path: str, modules: List[str]) -> None:
        # This method modifies imports in the generated files to be relative:
        with open(path, "rt") as file_in:
            file_content = file_in.readlines()

        with open(path, "wt") as file_out:
            for line in file_content:
                for module in modules:
                    line = line.replace("import {}_pb2 ".format(module), "from . import {}_pb2 ".format(module))
                file_out.write(line)

    def _ensure_protoc_version(self) -> None:
        """Checks that installed protoc has sufficient version."""
        protoc_args = [self._protoc_path, "--version"]
        version_stdout = subprocess.run(protoc_args, stdout=subprocess.PIPE, check=True).stdout

        # Convert bytestring to string and split into words.
        # b'libprotoc 3.7.1\n' => ['libprotoc', '3.7.1']
        version = version_stdout.decode("utf-8").strip().split()

        # Expected output is like 'libprotoc 3.7.1'
        if len(version) != 2 or not re.match(r"\d+.\d+.\d+", version[1]):
            raise RuntimeError(f"Unexpected output on resolving protoc version: {version}")

        # "3.7.1" => (3, 7, 1)
        parsed_version = tuple(map(int, version[1].split(".")))
        if parsed_version < PROTOC_MIN_VERSION:
            raise RuntimeError(
                f"Installed version of protoc is too old: {parsed_version}, install at least {PROTOC_MIN_VERSION}"
            )

    def compile(self, path_in: str, path_out: str, include: str = None) -> None:
        """Compiles .proto files from the `path_in` to `path_out` folder.

        Parameters
        ----------
        path_in: str
            Input folder.
        path_out: str
            Output folder.
        include: Optional[str]
            Includes.
        """
        os.makedirs(path_out)

        init_file_path = os.path.join(path_out, "__init__.py")
        open(init_file_path, "a").close()

        if include:
            protoc_args = [
                self._protoc_path,
                "--proto_path={}".format(path_in),
                "--proto_path={}".format(include),
                "--python_out={}".format(path_out),
            ]
        else:
            protoc_args = [self._protoc_path, "--proto_path={}".format(path_in), "--python_out={}".format(path_out)]

        proto_files = _find_proto_files(path_in)
        if include:
            proto_files.extend(_find_proto_files(include))
        protoc_args.extend(proto_files)
        protoc_process = subprocess.Popen(protoc_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        code = protoc_process.wait()
        (_, stderr) = protoc_process.communicate()
        if code == 0:
            logger.debug("Proto files were compiled successfully: %s", proto_files)
        else:
            logger.error("Error acquired while compiling files: %s. Files: %s.", stderr.decode("utf-8"), proto_files)

        modules = [proto_path.replace(".proto", "") for proto_path in proto_files]
        for file in filter(lambda f: f.endswith(".py"), os.listdir(path_out)):
            self._modify_file("{}/{}".format(path_out, file), modules)



"""Module with the Common Cryptography-assotiated Utils.

This module uses libsodium as a backend."""
from typing import Optional

from pysodium import (
    crypto_sign_keypair,
    crypto_sign_detached,
    crypto_sign_verify_detached,
    crypto_hash_sha256,
    crypto_hash_sha256_BYTES,
    crypto_sign_SECRETKEYBYTES,
    crypto_sign_PUBLICKEYBYTES,
    crypto_sign_BYTES,
    crypto_sign_init,
    crypto_sign_update,
    crypto_sign_final_create,
    crypto_sign_final_verify,
    CryptoSignState
)

HASH_BYTES_LEN = crypto_hash_sha256_BYTES
PUBLIC_KEY_BYTES_LEN = crypto_sign_PUBLICKEYBYTES
SECRET_KEY_BYTES_LEN = crypto_sign_SECRETKEYBYTES
SIGNATURE_BYTES_LEN = crypto_sign_BYTES


# In this module classes are used as a storage and also contain verification.
# pylint: disable=too-few-public-methods


class _FixedByteArray:
    """Base class for types which store a bytes sequence of a fixed length."""

    def __init__(self, data: bytes, expected_len: int):
        if len(data) != expected_len:
            raise ValueError(f"Incorrect data length: expected {expected_len}, got {len(data)}.")

        self.value = data

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _FixedByteArray):
            return False
        return self.value == other.value

    def __str__(self) -> str:
        return self.hex()

    def hex(self) -> str:
        """Returns a hex representation of the value."""
        return self.value.hex()


class Hash(_FixedByteArray):
    """Representation of a SHA-256 hash."""

    def __init__(self, hash_bytes: bytes):
        super().__init__(hash_bytes, HASH_BYTES_LEN)

    @classmethod
    def hash_data(cls, data: Optional[bytes]) -> "Hash":
        """Calculates a hash of the provided bytes sequence and returns a Hash object.

        If `None` is provided, a hash of the empty sequence will be returned."""
        if data is not None:
            hash_bytes = crypto_hash_sha256(data)
        else:
            hash_bytes = crypto_hash_sha256(bytes())
        return cls(hash_bytes)


class PublicKey(_FixedByteArray):
    """Representation of a Curve25519 Public Key."""

    def __init__(self, key: bytes):
        super().__init__(key, PUBLIC_KEY_BYTES_LEN)


class SecretKey(_FixedByteArray):
    """Representation of a Curve25519 Secret Key."""

    def __init__(self, key: bytes):
        super().__init__(key, SECRET_KEY_BYTES_LEN)


class KeyPair:
    """Representation of a Curve25519 keypair."""

    def __init__(self, public_key: PublicKey, secret_key: SecretKey):
        # Check that public_key corresponds to the secret_key.
        # Since we use only the libsodium backend, it is normal to make this
        # check as presented.
        # libsodium secret key contains a public key inside.
        if secret_key.value[PUBLIC_KEY_BYTES_LEN:] != public_key.value:
            raise ValueError("Public key doesn't correspond to the secret key.")

        self.public_key = public_key
        self.secret_key = secret_key

    @classmethod
    def generate(cls) -> "KeyPair":
        """Generates a new random keypair"""
        public_key, secret_key = crypto_sign_keypair()
        return cls(PublicKey(public_key), SecretKey(secret_key))


class Signature(_FixedByteArray):
    """Representation of a Curve25519 signature"""

    def __init__(self, signature: bytes):
        super().__init__(signature, SIGNATURE_BYTES_LEN)

    @classmethod
    def sign(cls, data: bytes, key: SecretKey) -> "Signature":
        """Signs the provided bytes sequence with the provided secret key."""

        signature = crypto_sign_detached(data, key.value)

        return Signature(signature)

    def verify(self, data: bytes, key: PublicKey) -> bool:
        """Verifies the signature against the provided data and the public key."""

        try:
            crypto_sign_verify_detached(self.value, data, key.value)
            return True
        except ValueError:
            # ValueError is raised if verification fails.
            return False


class SignStream(object):
    """This structure provides a possibility to create and/or verify
       digital signatures for a stream of data. If the data are split into several
       chunks, the indicated chunks are added to the system and when adding is
       complete, the data is signed."""

    def __init__(self, state: CryptoSignState):
        self.state = state

    @classmethod
    def init(cls) -> "SignStream":
        """Creates a new instance of `SignStream`."""

        state = crypto_sign_init()
        return SignStream(state)

    def update(self, data: bytes) -> "SignStream":
        """Adds a new `chunk` to the message that will eventually be signed and/or verified."""

        crypto_sign_update(self.state, data)
        return self

    def finalize(self, key: SecretKey) -> "Signature":
        """Computes and returns a signature for the previously supplied message using the given `key`."""

        signature = crypto_sign_final_create(self.state, key.value)
        return Signature(signature)

    def verify(self, signature: Signature, key: PublicKey) -> bool:
        """Verifies that `signature` is a valid signature for the previously supplied message using the given `key`."""

        try:
            crypto_sign_final_verify(self.state, signature.value, key.value)
            return True
        except ValueError:
            # ValueError is raised if verification fails.
            return False



"""Common errors that can occur during work with ListProofs."""
from enum import Enum, auto as enum_auto

# Methods are self-documenting here.
# pylint: disable=missing-docstring


class MalformedListProofError(Exception):
    """
    Error is raised if the proof is malformed.
    Every object of this class contains field `error_kind` which is a value of type ErrorKind.
    This value describes the kind of the occured error.
    See enum ErrorKind for details.
    """

    class ErrorKind(Enum):
        """
        Kind of the error. Possible variants:
          - UNEXPECTED_LEAF: Proof contains a hash in the place where a value was expected.
          - UNEXPECTED_BRANCH: Proof contains a hash in the position which is impossible according to the list length.
          - REDUNDANT_HASH: There are redundant hashes in the proof: the hash of the underlying list can be calculated
            without some present hashes.
          - MISSING_HASH: Proof does not contain necessary information to compute the hash of the underlying list.
          - NON_EMPTY_PROOF: Non-empty proof for an empty list.
          - DUPLICATE_KEY: Same key is used more than once in the proof.
        """

        UNEXPECTED_LEAF = enum_auto()
        UNEXPECTED_BRANCH = enum_auto()
        REDUNDANT_HASH = enum_auto()
        MISSING_HASH = enum_auto()
        NON_EMPTY_PROOF = enum_auto()
        PARSE_ERROR = enum_auto()
        DUPLICATE_KEY = enum_auto()

    def __init__(self, message: str, error_kind: "ErrorKind") -> None:
        super().__init__(message)

        self.error_kind = error_kind

    @classmethod
    def unexpected_leaf(cls) -> "MalformedListProofError":
        error_msg = "Unexpected leaf"
        error_kind = cls.ErrorKind.UNEXPECTED_LEAF

        return cls(error_msg, error_kind)

    @classmethod
    def unexpected_branch(cls) -> "MalformedListProofError":
        error_msg = "Unexpected branch"
        error_kind = cls.ErrorKind.UNEXPECTED_BRANCH

        return cls(error_msg, error_kind)

    @classmethod
    def redundant_hash(cls) -> "MalformedListProofError":
        error_msg = "Redundant hash"
        error_kind = cls.ErrorKind.REDUNDANT_HASH

        return cls(error_msg, error_kind)

    @classmethod
    def missing_hash(cls) -> "MalformedListProofError":
        error_msg = "Missing hash"
        error_kind = cls.ErrorKind.MISSING_HASH

        return cls(error_msg, error_kind)

    @classmethod
    def non_empty_proof(cls) -> "MalformedListProofError":
        error_msg = "Non-empty proof"
        error_kind = cls.ErrorKind.NON_EMPTY_PROOF

        return cls(error_msg, error_kind)

    @classmethod
    def parse_error(cls, message: str) -> "MalformedListProofError":
        error_msg = f"Parsing error: could not parse {message}"
        error_kind = cls.ErrorKind.PARSE_ERROR

        return cls(error_msg, error_kind)

    @classmethod
    def duplicate_key(cls) -> "MalformedListProofError":
        error_msg = "Duplicate key"
        error_kind = cls.ErrorKind.DUPLICATE_KEY

        return cls(error_msg, error_kind)


class ListProofVerificationError(Exception):
    """ Error raised when the provided root hash does not match the calculated one. """

    def __init__(self, provided_hash: bytes, calculated_hash: bytes) -> None:
        def _short_hash(hash_bytes: bytes) -> str:
            hash_str = hash_bytes.hex()
            return hash_str[:4] + ".." + hash_str[-4:]

        err_msg = "Unmatched root hash. Got: {}, expected {}"
        super().__init__(err_msg.format(_short_hash(provided_hash), _short_hash(calculated_hash)))
        self.provided_hash = provided_hash
        self.calculated_hash = calculated_hash



"""MapProof Module."""
from typing import Optional, Dict, Any, List, Iterator, Callable
from logging import getLogger

from exonum_client_cis.crypto import Hash
from .constants import KEY_SIZE
from .proof_path import ProofPath
from .errors import MalformedMapProofError
from .optional_entry import OptionalEntry
from .branch_node import BranchNode
from ..hasher import Hasher
from ..utils import is_field_hash, to_bytes

# pylint: disable=C0103
logger = getLogger(__name__)


class _MapProofEntry:
    def __init__(self, path: ProofPath, data_hash: Hash):
        self.path = path
        self.hash = data_hash

    def __repr__(self) -> str:
        return "Entry [path: {}, hash: {}]".format(self.path, self.hash)

    @staticmethod
    def parse(data: Dict[str, str]) -> "_MapProofEntry":
        """ Parses MapProofEntry from the provided dict. """

        if not isinstance(data.get("path"), str) or not is_field_hash(data, "hash"):
            err = MalformedMapProofError.malformed_entry(data)
            logger.warning(str(err))
            raise err

        path_bits = data["path"]
        path = ProofPath.parse(path_bits)

        data_hash = to_bytes(data["hash"])
        if data_hash is None:
            err = MalformedMapProofError.malformed_entry(data)
            logger.warning(str(err))
            raise err

        return _MapProofEntry(path, Hash(data_hash))


def collect(entries: List[_MapProofEntry]) -> Hash:
    """
    Computes the root hash of the Merkle Patricia tree backing the specified entries
    in the map view.
    The tree is not restored in full; instead, we add the paths to
    the tree in their lexicographic order (i.e., according to the `PartialOrd` implementation
    of `ProofPath`) and keep track of the rightmost nodes (the right contour) of the tree.
    It is easy to see that adding paths in the lexicographic order means that only
    the nodes in the right contour may be updated on each step. Further, on each step
    zero or more nodes are evicted from the contour, and a single new node is
    added to it.
    `entries` are assumed to be sorted by the path in the increasing order.
    """

    def common_prefix(left: ProofPath, right: ProofPath) -> ProofPath:
        return left.prefix(left.common_prefix_len(right))

    def hash_branch(left_child: _MapProofEntry, right_child: _MapProofEntry) -> Hash:
        branch = BranchNode()
        branch.set_child("left", left_child.path, left_child.hash)
        branch.set_child("right", right_child.path, right_child.hash)

        return branch.object_hash()

    def fold(contour: List[_MapProofEntry], last_prefix: ProofPath) -> Optional[ProofPath]:
        last_entry = contour.pop()
        penultimate_entry = contour.pop()

        contour.append(_MapProofEntry(path=last_prefix, data_hash=hash_branch(penultimate_entry, last_entry)))

        if len(contour) > 1:
            penultimate_entry = contour[len(contour) - 2]
            return common_prefix(penultimate_entry.path, last_prefix)

        return None

    if not entries:
        return Hash(Hasher.DEFAULT_HASH)

    if len(entries) == 1:
        if not entries[0].path.is_leaf():
            err = MalformedMapProofError.non_terminal_node(entries[0].path)
            logger.warning(str(err))
            raise err

        return Hasher.hash_single_entry_map(entries[0].path.as_bytes(), entries[0].hash)

    # There is more than 1 entry.

    # Contour of entries to be folded into the result hash:
    contour: List[_MapProofEntry] = []

    # Initical contour state:
    first_entry, second_entry = entries[0], entries[1]
    last_prefix = common_prefix(first_entry.path, second_entry.path)
    contour = [first_entry, second_entry]

    # Process the rest of the entries:
    for entry in entries[2:]:
        new_prefix = common_prefix(contour[-1].path, entry.path)

        # Fold contour from the latest added entry to the beginning.
        # At each iteration take two latest entries and attempt to fold them into one new entry:
        while len(contour) > 1 and len(new_prefix) < len(last_prefix):
            prefix = fold(contour, last_prefix)
            if prefix is not None:
                last_prefix = prefix

        contour.append(entry)
        last_prefix = new_prefix

    # All entries are processed. Fold the contour into the final hash:
    while len(contour) > 1:
        prefix = fold(contour, last_prefix)
        if prefix:
            last_prefix = prefix

    logger.debug("Successfully computed the root hash of the Merkle Patricia tree.")
    return contour[0].hash


class CheckedMapProof:
    """ Version of `MapProof` obtained after verification. """

    def __init__(self, entries: List[OptionalEntry], root_hash: Hash):
        self._entries = entries
        self._root_hash = root_hash

    def missing_keys(self) -> Iterator[OptionalEntry]:
        """ Retrieves entries that the proof shows as missing from the map. """
        return filter(lambda el: el.is_missing, self._entries)

    def entries(self) -> Iterator[OptionalEntry]:
        """ Retrieves entries that the proof shows as present in the map. """
        return filter(lambda el: not el.is_missing, self._entries)

    def all_entries(self) -> List[OptionalEntry]:
        """ Retrieves all entries in the proof. """
        return self._entries

    def root_hash(self) -> Hash:
        """ Returns a hash of the map for which this proof is constructed. """
        return self._root_hash


class MapProof:
    """
    Proof of existence for a map.

    This class is capable of parsing MapProof from Dict[Any, Any] and
    verifying the parsed proof.

    In most cases you do not need to create MapProof manually (use MapProofBuilder instead).

    However, if you understand that MapProofBuilder does not suit your needs, use the
    following workflow:

    >>> proof = get_proof_somehow()
    >>> # Assuming that the Protobuf files for the service are already loaded and compiled.
    >>> cryptocurrency_service_name = 'exonum-cryptocurrency-advanced:0.11.0'
    >>> cryptocurrency_module = ModuleManager.import_service_module(cryptocurrency_service_name, 'service')
    >>> cryptocurrency_decoder = MapProofBuilder.build_encoder_function(cryptocurrency_module.Wallet)
    >>> # Keys will be encoded to bytes using "bytes.fromhex",
    >>> # and values will be encoded using "Wallet" Protobuf structure.
    >>> parsed_proof = MapProof.parse(proof, lambda x: bytes.fromhex(x), cryptocurrency_decoder)
    >>> # CheckedMapProof will be returned.
    >>> result = parsed_proof.check()
    ```
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        entries: List[OptionalEntry],
        proof: List[_MapProofEntry],
        key_to_bytes: Callable[[Any], bytes],
        value_to_bytes: Callable[[Any], bytes],
        raw: bool = False,
    ):
        """
        Constructor of MapProof. Do not call it directly, use `MapProof.parse`.
        """
        self.entries = entries
        self.proof = proof
        self._key_to_bytes = key_to_bytes
        self._value_to_bytes = value_to_bytes
        self._raw = raw

    def __repr__(self) -> str:
        format_str = "MapProof [\n  Entries: {}\n  Proof: {}\n]\n"

        return format_str.format(self.entries, self.proof)

    @staticmethod
    def parse(
        data: Dict[str, Any],
        key_to_bytes: Callable[[Any], bytes],
        value_to_bytes: Callable[[Any], bytes],
        raw: bool = False,
    ) -> "MapProof":
        """
        Method to parse a proof.

        Example of the expected proof format:

        >>>
        {
          "entries": [
            {
              "key": "e610db75b0bbbd4c606c4f8ca3fca9f916e9c8ae9a93b5b767082172454344b3",
              "value": {
                "pub_key": {
                  "data": list(bytes.fromhex("e610db75b0bbbd4c606c4f8ca3fca9f916e9c8ae9a93b5b767082172454344b3"))
                },
                "name": "Alice1",
                "balance": 95,
                "history_len": 6,
                "history_hash": {
                  "data": list(bytes.fromhex("19faf859d7456907c76f085af5b7a2d7621d992617a349006c07720957d5d49d"))
                }
              }
            }
          ],
          "proof": [
            {
              "path": "010000100100100010100100110001100111011011011000010011011011111111110110001101001001011010111100"
                      "001010010011011010001110011001011001010101100100000101010010000001101011100000010011101101001011"
                      "1110010011011011101001110000111111000000011111100010001011010000",
              "hash": "dbeab4aa952e2c2cb3dc921aa42c9b508e2e5961cad2463f7203d228abc204c8"
            }
          ]
        }

        Parameters
        ----------
        data: Dict[str, Any]
            Name of the Protobuf structure to be used in the converted function.
        key_to_bytes: Callable[[Any], bytes]
            Function that will be used to convert keys to bytes.
        value_to_bytes: Callable[[Any], bytes]
            Function that will be used to convert values to bytes.
        raw: bool
            Denotes if type of the proof is `RawMapProof` (meaning that key should not be hashed and
            maps directly to the `ProofPath`).
            Defaults to `False`.

        Returns
        -------
        MapProof
            If parsing succeeds, MapProof object is returned.

        Raises
        ------
        MalformedMapProofError
            If provided raw MapProof is malformed and parsing fails, this exception
            rises.
        """

        if data.get("entries") is None or data.get("proof") is None:
            logger.warning("'entries' or 'proof' field is missing in the proof dictionary.")
            raise MalformedMapProofError.malformed_entry(data)

        entries: List[OptionalEntry] = [OptionalEntry.parse(raw_entry) for raw_entry in data["entries"]]
        proof: List[_MapProofEntry] = [_MapProofEntry.parse(raw_entry) for raw_entry in data["proof"]]

        map_proof = MapProof(entries, proof, key_to_bytes, value_to_bytes, raw)
        logger.debug("Successfully built MapProof from the given proof dictionary.")
        return map_proof

    @staticmethod
    def _check_proof(proof: List[_MapProofEntry]) -> None:
        for idx in range(1, len(proof)):
            prev_path, path = proof[idx - 1].path, proof[idx].path

            if prev_path < path:
                if path.starts_with(prev_path):
                    err = MalformedMapProofError.embedded_paths(prev_path, path)
                    logger.warning(str(err))
                    raise err
            elif prev_path == path:
                err = MalformedMapProofError.duplicate_path(path)
                logger.warning(str(err))
                raise err
            elif prev_path > path:
                err = MalformedMapProofError.invalid_ordering(prev_path, path)
                logger.warning(str(err))
                raise err
            else:
                assert False, "Incomparable keys in the proof"

    def check(self) -> CheckedMapProof:
        """
        Calculates the root hash from the parsed proof.

        Returns
        -------
        CheckedMapProof
            If no errors occur during the calculation, a CheckedMapProof entity is returned.

        Raises
        ------
        MalformedMapProofError
            If provided raw MapProof is malformed and parsing fails, this exception
            rises.
        """

        def kv_to_map_entry(key_value: OptionalEntry) -> _MapProofEntry:
            key_bytes = self._key_to_bytes(key_value.key)
            if self._raw:
                # For raw map proof keys aren't hashed.
                if len(key_bytes) != KEY_SIZE:
                    raise MalformedMapProofError.invalid_key_size(key_bytes)
                key = key_bytes
            else:
                # For usual proofs we need to hash the key.
                key = Hasher.hash_raw_data(key_bytes).value

            path = ProofPath.from_bytes(key)
            value_hash = Hasher.hash_leaf(self._value_to_bytes(key_value.value))

            return _MapProofEntry(path, value_hash)

        proof = self.proof[:]

        actual_entries = filter(lambda el: not el.is_missing, self.entries)

        proof += map(kv_to_map_entry, actual_entries)

        proof.sort(key=lambda el: el.path)

        self._check_proof(proof)

        result = collect(proof)

        return CheckedMapProof(self.entries, Hasher.hash_map_node(result))



"""Module capable of loading the Protobuf-generated modules."""
from typing import Any
import importlib
import re


class ModuleManager:
    """ModuleManager class provides an interface for importing modules generated from the previously downloaded
    Protobuf sources.

    It is supposed that you call those methods only after downloading the corresponding module via ProtobufLoader.
    Otherwise an error will be raised.

    Example usage:

    >>> with client.protobuf_loader() as loader:
    >>>     loader.load_main_proto_files()
    >>>     loader.load_service_proto_files(0, "exonum-supervisor:0.13.0-rc.2")
    >>>     main_module = ModuleManager.import_main_module("runtime")
    >>>     service_module = ModuleManager.import_service_module("exonum-supervisor:0.13.0-rc.2", "service")
    """

    @staticmethod
    def import_main_module(module_name: str) -> Any:
        """Imports the main (used by the Exonum core) module, e.g. "consensus", "runtime", etc."""
        module = importlib.import_module("exonum_modules.main.{}_pb2".format(module_name))

        return module

    @staticmethod
    def import_service_module(artifact_name: str, module_name: str) -> Any:
        """Imports the service (corresponding to some artifact) module."""
        artifact_module_name = re.sub(r"[-. :/]", "_", artifact_name)
        module = importlib.import_module("exonum_modules.{}.{}_pb2".format(artifact_module_name, module_name))

        return module



"""
Exonum API Module.

This module provides 4 classes classes:
  - Api: a class with basic REST functionality.
  - ProtobufApi: a class that implements ProtobufProviderInterface interface.
  - PublicApi: a subclass of class Api that provides methods to interact with public API of an Exonum node.
  - PrivateApi: a subclass of class Api that provides methods to interact with private API of an Exonum node.
  - ServiceApi: a class that provides methods to interact with node services.
"""
from typing import Optional, Any, Union, List, Dict, Iterable

import json
from logging import getLogger
import requests

from .protobuf_loader import ProtoFile, ProtobufProviderInterface
from .message import ExonumMessage

# pylint: disable=C0103
logger = getLogger(__name__)


class Api:
    """Api class provides basic REST functionality."""

    # constants
    RUST_RUNTIME_ID = 0

    def __init__(self, hostname: str, port: int, schema: str):
        """
        Constructor of Api.

        Parameters
        ----------
        hostname: str
            Examples: '127.0.0.1', 'www.some_node.com'.
        port: int
            API port of an Exonum node.
        schema: str
             Communication protocol: 'https' or 'http'.
        """
        self.schema = schema
        self.hostname = hostname
        self.port = port
        # Example of a formatted prefix: "https://127.0.0.1:8000"
        self.endpoint_prefix = "{}://{}:{}/api".format(self.schema, hostname, port)

    @staticmethod
    def get(url: str, params: Optional[Dict[Any, Any]] = None) -> requests.Response:
        """Internal wrapper over requests.get"""
        return requests.get(url, params=params)

    @staticmethod
    def post(url: str, data: str, headers: Dict[str, str]) -> requests.Response:
        """Internal wrapper over requests.post"""
        return requests.post(url, data=data, headers=headers)


class ProtobufApi(Api, ProtobufProviderInterface):
    """ProtobufApi class implements ProtobufProviderInterface interface."""

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self._rust_runtime_url = self.endpoint_prefix + "/runtimes/rust/{}"

    def _get_proto_sources(self, params: Optional[Dict[str, str]] = None) -> List[ProtoFile]:
        """Retrieves protobuf sources."""
        proto_sources_endpoint = self._rust_runtime_url.format("proto-sources")
        response = self.get(proto_sources_endpoint, params=params)
        if response.status_code != 200 or "application/json" not in response.headers["content-type"]:
            logger.critical(
                "Unsuccessfully attempted to retrieve Protobuf sources.\n" "Status code: %s,\n" "body:\n%s",
                response.status_code,
                response.content,
            )
            raise RuntimeError("Unsuccessfully attempted to retrieve Protobuf sources: {!r}".format(response.content))
        logger.debug("Protobuf sources retrieved successfully.")

        proto_files = [
            ProtoFile(name=proto_file["name"], content=proto_file["content"]) for proto_file in response.json()
        ]

        return proto_files

    def get_main_proto_sources(self) -> List[ProtoFile]:
        """Performs a GET request to the `proto-sources` Exonum endpoint."""
        params = {"type": "core"}
        return self._get_proto_sources(params)

    def get_proto_sources_for_artifact(self, runtime_id: int, artifact_name: str) -> List[ProtoFile]:
        """Raise an exception if runtime ID is not equal to the rust runtime ID."""
        if runtime_id != self.RUST_RUNTIME_ID:
            err_msg = f"Provided runtime ID: {runtime_id} is not equal to Rust runtime ID: {self.RUST_RUNTIME_ID}."
            logger.critical(err_msg)
            raise RuntimeError(err_msg)
        # Performs a GET request to the `proto-sources` Exonum endpoint with a provided artifact name:
        params = { "artifact": artifact_name }

        return self._get_proto_sources(params)


class PublicApi(Api):
    """PublicApi class provides methods to interact with the public API of an Exonum node."""

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self._tx_url = self.endpoint_prefix + "/explorer/v1/transactions"
        self._block_url = self.endpoint_prefix + "/explorer/v1/block"
        self._blocks_url = self.endpoint_prefix + "/explorer/v1/blocks"
        self._system_url = self.endpoint_prefix + "/system/v1/{}"

    def get_block(self, height: int) -> requests.Response:
        """
        Gets a block at the provided height.

        Example:

        >>> public_api = PublicApi("127.0.0.1", 80, "http")
        >>> block = public_api.get_block(2).json()
        >>> print(json.dumps(block, indent=2))
        {
          "proposer_id": 0,
          "height": 2,
          "tx_count": 0,
          "prev_hash": "e686088d5323e51c096b42126a65fff59363c740ad0d8260c6c03c2e0c40ecdd",
          "tx_hash": "c6c0aa07f27493d2f2e5cff56c890a353a20086d6c25ec825128e12ae752b2d9",
          "state_hash": "e552443214f22721d007f1eef03f5e4d2483c31a439043eb32cd7b1faeef354f",
          "precommits": [
            "0a5c2...0603"
          ],
          "txs": [],
          "time": "2019-09-12T09:50:49.390408335Z"
        }

        Parameters
        ----------
        height: int
            A height of the required block.

        Returns
        -------
        block_response: requests.Response
            Result of an API call.
            If it is successful, a JSON representation of the block will be in the response.
        """
        return self.get(self._block_url, params={"height": height})

    # pylint: disable=too-many-arguments
    def get_blocks(
        self,
        count: int,
        earliest: Optional[int] = None,
        latest: Optional[int] = None,
        add_precommits: bool = False,
        skip_empty_blocks: bool = False,
        add_blocks_time: bool = False,
    ) -> requests.Response:
        """
        Gets a range of blocks.

        Blocks will be returned in a reversed order starting from the latest to the `latest - count + 1`.
        See the `latest` parameter description for details.

        Parameters
        ----------
        count: int
            Amount of blocks. Should not be greater than Exonum's parameter MAX_BLOCKS_PER_REQUEST
        earliest: Optional[int]
            If not provided, it is considered to be the height of the earliest block in the blockchain.
            Otherwise, a provided value will be used.
        latest: Optional[int]
            If not provided, it is considered to be the height of the latest block in the blockchain.
            Otherwise, a provided value will be used.
        add_precommits: bool
            If True, precommits will also be taken into account.
        skip_empty_blocks: bool
            If True, only non-empty blocks will be returned. By default it is False.
        add_blocks_time: bool
            If True, then the returned `times` field of BlockRange will contain a median time from the
            corresponding block precommits.

        Returns
        -------
        blocks_range_response: requests.Response
            Result of an API call.
            If it is successful, a JSON representation of the block range will be in the response.
        """
        params: Dict[str, Union[int, str]] = dict()
        params["count"] = count

        if earliest:
            params["earliest"] = earliest
        if latest:
            params["latest"] = latest
        if add_precommits:
            params["add_precommits"] = "true"
        if skip_empty_blocks:
            params["skip_empty_blocks"] = "true"
        if add_blocks_time:
            params["add_blocks_time"] = "true"

        return self.get(self._blocks_url, params=params)

    def get_tx_info(self, tx_hash: str) -> requests.Response:
        """
        Gets information about the transaction with the provided hash.

        Example:

        >>> public_api = PublicApi("127.0.0.1", 80, "http")
        >>> tx_info = public_api.get_tx_info(tx_hash).json()
        >>> print(json.dumps(tx_info, indent=2))
        {
          'type': 'committed',
          'content': '0a11...660d',
          'location': {
            'block_height': 58224,
            'position_in_block': 1
          },
          'location_proof': {
            'proof': [
              {
                'index': 0,
                'height': 1,
                'hash': '14637aa10b700cebfbc23d45395e8677d1fe1914d2e7f50d38cf1b73cfba1702'
              }
            ],
            'entries': [
              [1, 'e2d9ba5e8e104d65be8d3af7c26e5abea8f27da280cea110a80c9ab4f4d2a10c']
            ],
            'length': 2
          },
          'status': {
            'type': 'success'
          },
          'time': '2019-09-12T13:08:10.528537286Z'
        }

        Parameters
        ----------
        tx_hash: str
            A hexadecimal representation of the transaction hash.

        Returns
        -------
        block_response: requests.Response
            Result of an API call.
            If it is successful, a JSON representation of the transaction info will be in the response.
        """
        return self.get(self._tx_url, params={"hash": tx_hash})

    def available_services(self) -> requests.Response:
        """
        Gets a list of available services from Exonum.

        Example:

        >>> public_api = PublicApi("127.0.0.1", 80, "http")
        >>> available_services = public_api.available_services().json()
        >>> print(json.dumps(available_services, indent=2))
        {
          "artifacts": [
            {
              "runtime_id": 0,
              "name": "exonum-supervisor",
              "version": "0.13.0-rc.2"
            }
          ],
          "services": [
            {
              "spec": {
                "id": 0,
                "name": "supervisor",
                "artifact": {
                  "runtime_id": 0,
                  "name": "exonum-supervisor",
                  "version": "0.13.0-rc.2"
                }
              },
              "status": "Active"
            }
          ]
        }
        """
        return self.get(self._system_url.format("services"))

    def get_instance_id_by_name(self, name: str) -> Optional[int]:
        """
        Gets an ID of the service instance with the provided name.

        Example:

        >>> public_api = PublicApi("127.0.0.1", 80, "http")
        >>> id = public_api.get_instance_id_by_name("cryptocurrency")
        >>> id
        42

        Parameters
        ----------
        name: Name of the service.

        Returns
        -------
        result: Optional[int]
            ID of the instance (int) if the instance with given name exists, otherwise None is returned.

        Raises
        ------
        RuntimeError
            An error will be raised if a response code is not 200.
        """
        response = self.available_services()

        if response.status_code != 200:
            raise RuntimeError("Couldn't get info about available services")

        available_services = response.json()

        for state in available_services["services"]:
            service = state["spec"]
            if service["name"] == name:
                return service["id"]

        return None

    def send_transaction(self, message: ExonumMessage) -> requests.Response:
        """
        Sends a transaction into an Exonum node via REST API.

        Example:

        >>> public_api = PublicApi("127.0.0.1", 80, "http")
        >>> response = public_api.send_transaction(message)
        >>> response.json()
        {'tx_hash': '713de312f48fe15559c0d4f7fb3f274dfbd3893a8a80d9f4224e97248f0e314e'}

        Parameters
        ----------
        message: ExonumMessage
            Prepared and signed an Exonum message.

        Returns
        -------
        result: requests.Response
            Result of the POST request.
            If a transaction is correct and it is accepted, it will contain a JSON with a hash of the transaction.
        """
        response = self.post(self._tx_url, data=message.pack_into_json(), headers={"content-type": "application/json"})
        return response

    def send_transactions(self, messages: Iterable[ExonumMessage]) -> List[requests.Response]:
        """
        Same as send_transaction, but for any iterable object over ExonumMessage.

        Parameters
        ----------
        messages: Iterable[ExonumMessage]
            A sequence of messages to send.

        Returns
        -------
        results: List[requests.Response]
            A list of responses for each sent transaction.
        """
        return [self.send_transaction(message) for message in messages]

    def health_info(self) -> requests.Response:
        """ Performs a GET request to the healthcheck Exonum endpoint. """
        return self.get(self._system_url.format("healthcheck"))

    def stats(self) -> requests.Response:
        """ Performs a GET request to the stats Exonum endpoint. """
        return self.get(self._system_url.format("stats"))

    def user_agent(self) -> requests.Response:
        """ Performs a GET request to the user_agent Exonum endpoint. """
        return self.get(self._system_url.format("user_agent"))


class PrivateApi(Api):
    """PrivateApi class provides methods to interact with the private API of an Exonum node."""

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self._system_url = self.endpoint_prefix + "/system/v1/{}"

    def add_peer(self, address: str, public_key: str) -> requests.Response:
        """
        Performs a POST request to the '{system_base_path}/peers' endpoint in order to add a new peer to the node.

        Parameters
        ----------
        address: IP address of the node which the present node should connect to.
        public_key: public key of the node which the present node should connect to.

        Returns
        -------
        response: requests.Response
            Result of an API call.
        """
        data = json.dumps({"address": address, "public_key": public_key})
        response = self.post(self._system_url.format("peers"), data=data, headers={"content-type": "application/json"})
        return response

    def get_peers(self) -> requests.Response:
        """
        Performs a GET request to the '{system_base_path}/peers' to get a list of peers of the node.

        Example:

        >>> private_api = PrivateApi("127.0.0.1", 81, "http")
        >>> peers = private_api.get_peers().json()
        >>> print(json.dumps(peers, indent=2))
        {
          "incoming_connections": [{
            "address": "127.0.0.1:57671",
            "public_key": "8a17bdfe42c10abdb7f27b5648691db3338400c27812e847e02eb7193ad490f2"
          }],
          "outgoing_connections": {
            "127.0.0.1:6334": {
              "public_key": "dcb46dceaeb7d0eab7b6ed000f317f2ab9f7c8423ec9a6a602d81c0979e1333a",
              "state": {
                "type": "Active"
              }
            },
            "127.0.0.1:6335": {
              "public_key": "dcb46dceaeb7d0eab7b6ed000f317f2ab9f7c8423ec9a6a602d81c0979e1333a",
              "state": {
                "delay": 4000,
                "type": "Reconnect"
              }
            },
            "127.0.0.1:6336": {
              "public_key": null,
              "state": {
                "type": "Active"
              }
            },
            "127.0.0.1:6337": {
              "public_key": null,
              "state": {
                "delay": 4000,
                "type": "Reconnect"
              }
            }
          }
        }

        Returns
        -------
        response: requests.Response
            Result of an API call.
            If it is successful, a list of peers will be returned.
        """
        return self.get(self._system_url.format("peers"))

    def set_consensus_interaction(self, enabled: bool = True) -> requests.Response:
        """
        Performs a POST request to the '{system_base_path}/consensus_enabled'
        to switch consensus interaction of the node on or off.

        Parameters
        ----------
        enabled: flag that switches consensus interaction of the node on (True) or off (False).

        Returns
        -------
        response: requests.Response
            Result of an API call.
        """
        data = json.dumps({"enabled": enabled})
        return self.post(
            self._system_url.format("consensus_enabled"), data=data, headers={"content-type": "application/json"}
        )

    def get_consensus_interaction(self) -> requests.Response:
        """
        Performs a GET request to the '{system_base_path}/consensus_enabled' to get boolean value that states
        if the node participates in consensus.

        Example:

        >>> private_api = PrivateApi("127.0.0.1", 81, "http")
        >>> private_api.get_consensus_interaction().json()
        True

        Returns
        -------
        response: requests.Response
            Result of an API call.
        """
        return self.get(self._system_url.format("consensus_enabled"))

    def get_network_info(self) -> requests.Response:
        """
        Performs a GET request to the '{system_base_path}/network'
        to get info about the serialization protocol and the services functioning in the network.

        Example:

        >>> private_api = PrivateApi("127.0.0.1", 81, "http")
        >>> network_info = private_api.get_network_info().json()
        >>> print(json.dumps(network_info, indent=2))
        {
          "core_version": "0.10.2"
        }

        Returns
        -------
        response: requests.Response
            Result of an API call.
        """
        return self.get(self._system_url.format("network"))

    def shutdown(self) -> requests.Response:
        """
        Performs a POST request to the '{system_base_path}/shutdown' to stop the node.
        After receiving this request, the node stops processing transactions, participating in consensus and
        terminates after all messages in the event queue are processed.

        Returns
        -------
        response: requests.Response
            Result of an API call.
        """
        data = json.dumps(None)
        return self.post(self._system_url.format("shutdown"), data=data, headers={"content-type": "application/json"})


class ServiceApi(Api):
    """ServiceApi class provides methods to interact with service API."""

    def __init__(self, service_name: str, *args: Any, **kwargs: Any):
        """
        Constructor of ServiceApi.

        Parameters
        ----------
        service_name: str
            Name of a service.
        *args, **kwargs:
            Arguments for Api class constructor.
        """
        super().__init__(*args, **kwargs)

        self._api_url = f"{self.schema}://{self.hostname}:{self.port}/api/services/{service_name}/"

    def service_endpoint(self, sub_uri: str) -> str:
        """
        Creates a service endpoint for a given sub-uri.

        Example:

        >>> service_api = ServiceApi("supervisor", "127.0.0.1", "8080", "http")
        >>> service_api.service_endpoint("deploy-artifact")
        http://127.0.0.1:8081/api/services/supervisor/deploy-artifact

        Parameters
        ----------
        sub_uri: str
            Additional part of a URL to be added to the endpoint, e.g. "some/sub/uri?parameter=value"
        private: bool
            Denotes if a private port should be used. Defaults to False.

        Returns
        -------
        url: str
            Returns a service REST API URL based on the provided parameters.
        """
        return self._api_url + sub_uri

    def get_service(self, sub_uri: str) -> requests.Response:
        """
        Performs a GET request to the endpoint generated by the `service_endpoint` method.

        Parameters are the same as in `service_endpoint`.

        Returns
        -------
        response: requests.Response
            Result of an API call.
        """
        return self.get(self.service_endpoint(sub_uri))

    def post_service(self, sub_uri: str, data: str) -> requests.Response:
        """
        Performs a POST request to the endpoint generated by the `service_endpoint` method.

        Parameters are the same as in `service_endpoint` except for `data`.
        `data` is expected to be a serialized JSON value.

        Returns
        -------
        response: requests.Response
            Result of an API call.
        """
        json_headers = {"content-type": "application/json"}
        return self.post(self.service_endpoint(sub_uri), data=data, headers=json_headers)



"""Module Containing the ProtobufLoader Class.

ProtobufLoader is capable of downloading Protobuf sources from Exonum."""
from typing import List, Optional, Any, NamedTuple
from logging import getLogger
import shutil
import sys
import os
import tempfile
import re

from .protoc import Protoc

# pylint: disable=C0103
logger = getLogger(__name__)

PYTHON_RUNTIME = 2


class ProtoFile(NamedTuple):
    """Structure that represents a proto file."""

    name: str
    content: str


class ProtobufProviderInterface:
    """Interface for Protobuf sources provider."""

    def get_main_proto_sources(self) -> List[ProtoFile]:
        """Gets the Exonum core proto sources."""
        raise NotImplementedError

    def get_proto_sources_for_artifact(
        self, runtime_id: int, artifact_name: str
    ) -> List[ProtoFile]:
        """Gets the Exonum core proto sources."""
        raise NotImplementedError


class ProtobufLoader:
    """ProtobufLoader is a class capable of loading and compiling Protobuf sources from Exonum.

    This class is a Singleton, meaning that only one entity of that class is created at a time.

    Example workflow:

    >>> with client.protobuf_loader() as loader:
    >>>    loader.load_main_proto_files()
    >>>    loader.load_service_proto_files(0, "exonum-supervisor:0.13.0-rc.2")

    Code above will initialize loader, download core Exonum proto files and proto files for the Supervisor service.
    The code will compile the files into the Python modules. After that you will be able to load those modules
    via ModuleManager.

    Please note that it is recommended to create a ProtobufLoader object via the context manager.
    Otherwise you will have to call `initialize` and `deinitialize` methods manually:

    >>> loader = client.protobuf_loader()
    >>> loader.initialize()
    >>> ... # Some code
    >>> loader.deinitialize()

    If you forget to call `deinitialize` (or if the code exits earlier, for example because of unhandled exception),
    the recourses created in the temp folder (which may differ depending on your OS) will not be removed.

    Creating more than one entity at a time will result in retrieving the same object:

    >>> with client.protobuf_loader() as loader_1:
    >>>     with client.protobuf_loader() as loader_2:
    >>>         assert loader_1 == loader_2

    This may be useful if you have several modules that should work with ProtobufLoader:

    >>> # main.py
    >>> loader = ProtobufLoader(client)
    >>> loader.initialize()
    >>> loader.load_main_proto_files()
    >>> ...
    >>> loader.deinitialize()

    >>> # module_a.py
    >>> loader = ProtobufLoader() # Since loader is already initialized with the client, you do not have to provide it.
    >>> loader.load_service_proto_files(runtime_a, service_a)

    >>> # module_b.py
    >>> loader = ProtobufLoader()
    >>> loader.load_service_proto_files(runtime_b, service_b)

    However, if you try to create the second loader, different from the first one, from the client,
    ValueError will be raised.
    """

    _entity = None
    _reference_count = 0

    def __new__(cls, *_args: Any) -> "ProtobufLoader":
        # Check if the entity is already created (and thus no new object should be created):
        if ProtobufLoader._entity is not None:
            return ProtobufLoader._entity

        # Create a new object:
        return super().__new__(cls)

    def __init__(self, client: Optional[ProtobufProviderInterface] = None):
        # Check that the client is the same as expected:
        if ProtobufLoader._entity is not None:
            if client is not None and client != ProtobufLoader._entity.client:
                err_msg = (
                    f"Attempt to create ProtobufLoader entity with a different client:\n"
                    f"used client:\n{ProtobufLoader._entity.client}\n"
                    f"provided client:\n{client}\n"
                )
                logger.critical(err_msg)
                raise ValueError(err_msg)
            return

        if client is None:
            err_msg = "Client is expected to be set for the initial object creation."
            logger.critical(err_msg)
            raise ValueError(err_msg)

        ProtobufLoader._entity = self
        self.client = client
        self.protoc = Protoc()
        self._proto_dir: Optional[str] = None

    def __enter__(self) -> "ProtobufLoader":
        self.initialize()

        return self

    def __exit__(self, exc_type: Optional[type], exc_value: Optional[Any], exc_traceback: Optional[object]) -> None:
        self.deinitialize()

    def initialize(self) -> None:
        """Performs an initialization process."""

        # Update the reference counter:
        ProtobufLoader._reference_count += 1
        logger.debug("Current ProtobufLoader reference count: %s.", ProtobufLoader._reference_count)
        if ProtobufLoader._reference_count > 1:
            # If this is a second (third, etc) entity, everything is already initialized:
            return

        # Create a directory for temporary files:
        self._proto_dir = tempfile.mkdtemp(prefix="exonum_client_cis_")

        # Create a folder for Python files output:
        python_modules_path = os.path.join(self._proto_dir, "exonum_modules")
        os.makedirs(python_modules_path)

        # Create __init__ file in the exonum_modules directory:
        init_file_path = os.path.join(python_modules_path, "__init__.py")
        open(init_file_path, "a").close()

        # Add a directory with exonum_modules into the Python path:
        sys.path.append(self._proto_dir)

        logger.debug("Successfully initialized ProtobufLoader for client:\n%s\n", self.client)

    def deinitialize(self) -> None:
        """Performs a deinitialization process."""
        if self._proto_dir is None:
            err_msg = "Attempt to deinitialize uninitialized ProtobufLoader."
            logger.critical(err_msg)
            raise RuntimeError(err_msg)

        # Decrement the reference counter:
        ProtobufLoader._reference_count -= 1
        logger.debug("Current ProtobufLoader reference count: %s.", ProtobufLoader._reference_count)

        # If there is at least one reference, nothing should be done:
        if ProtobufLoader._reference_count > 0:
            return

        # Mark entity as removed:
        ProtobufLoader._entity = None

        # Remove the generated temporary directory:
        sys.path.remove(self._proto_dir)
        shutil.rmtree(self._proto_dir)

        # Unload any previously loaded protobuf modules
        loaded_modules = list(sys.modules.keys())
        for module in loaded_modules:
            if module.startswith("exonum_modules"):
                del sys.modules[module]

        logger.debug("Successfully deinitialized ProtobufLoader.")

    @staticmethod
    def _save_proto_file(path: str, file_content: str) -> None:
        with open(path, "wt") as file_out:
            file_out.write(file_content)

    def _save_files(self, path: str, files: List[ProtoFile]) -> None:
        os.makedirs(path)
        for proto_file in files:
            file_path = os.path.join(path, proto_file.name)
            self._save_proto_file(file_path, proto_file.content)

    def load_main_proto_files(self) -> None:
        """Loads and compiles the main Exonum proto files."""
        if self._proto_dir is None:
            err_msg = "Attempt to use uninitialized ProtobufLoader."
            logger.critical(err_msg)
            raise RuntimeError(err_msg)

        # This method is not intended to be used by end users, but it is OK to call it here.
        # pylint: disable=protected-access
        proto_contents = self.client.get_main_proto_sources()

        # Save proto_sources in the proto/main directory:
        main_dir = os.path.join(self._proto_dir, "proto", "main")
        self._save_files(main_dir, proto_contents)

        # Call protoc to compile proto sources:
        proto_dir = os.path.join(self._proto_dir, "exonum_modules", "main")
        self.protoc.compile(main_dir, proto_dir)

    def load_service_proto_files(self, runtime_id: int, artifact_name: str) -> None:
        """Loads and compiles proto files for a service."""
        if self._proto_dir is None:
            logger.critical("Attempt to use unititialized ProtobufLoader.")
            raise RuntimeError("Attempt to use unititialized ProtobufLoader")

        # This method is not intended to be used by end users, but it is OK to call it here.
        # pylint: disable=protected-access
        proto_contents = self.client.get_proto_sources_for_artifact(runtime_id, artifact_name)

        # Save proto_sources in proto/artifact_name directory:
        service_module_name = re.sub(r"[-. :/]", "_", artifact_name)
        service_dir = os.path.join(self._proto_dir, "proto", service_module_name)
        self._save_files(service_dir, proto_contents)

        # Call protoc to compile proto sources:
        main_dir = os.path.join(self._proto_dir, "proto", "main")
        proto_dir = os.path.join(self._proto_dir, "exonum_modules", service_module_name)
        if runtime_id != PYTHON_RUNTIME:
            self.protoc.compile(service_dir, proto_dir, include=main_dir)
        else:
            # Python services do not rely on the `includes` from the exonum core.
            self.protoc.compile(service_dir, proto_dir)



"""Tools for Creating Value Binary Encoders."""
from typing import Callable, Dict, Any
import json

import base64
import copy

from google.protobuf import json_format


# TODO This custom is required because of a bug in the Python Protobuf implementation:
# https://github.com/protocolbuffers/protobuf/issues/6602
# This workaround is sort of a kludge, and it may lead to an incorrect behavior for int8/uint8 arrays.
# But currently Exonum does not have them, so it seems OK. When the issue is fixed,
# this code will be removed.
def _encode_byte_lists_to_base64(obj: Dict[Any, Any]) -> Dict[Any, Any]:
    def _visit(entries: Dict[Any, Any]) -> None:
        for key, value in entries.items():
            if isinstance(value, dict):
                _visit(value)
            elif isinstance(value, list) and all(map(lambda x: 0 <= x <= 255, value)):
                obj_raw = bytes(value)
                entries[key] = str(base64.b64encode(obj_raw), "utf-8")

    obj_copy = copy.deepcopy(obj)

    _visit(obj_copy)

    return obj_copy


def build_encoder_function(encoder_class: type) -> Callable[[Dict[Any, Any]], bytes]:
    """
    With this function you can create an encoder function for a value.

    It creates a Protobuf converter from a value (as a dict) to bytes which can be used either
    in MapProof or in ListProof.

    Parameters
    ----------
    encoder_class: type
        Class of the Protobuf message that will be used to convert data to bytes.

    Returns
    ------
    Callable[[Dict[Any, Any]], bytes]
        A converter function from a value to bytes.
    """

    def func(data: Dict[Any, Any]) -> bytes:
        # TODO temporary workaround for the problem described above.
        preencoded_data = _encode_byte_lists_to_base64(data)

        data_json = json.dumps(preencoded_data)
        protobuf_obj = encoder_class()

        return json_format.Parse(data_json, protobuf_obj, True).SerializeToString()

    return func



"""Module with OptionalEntry of MapProof."""
from typing import Optional, Any, Dict
from logging import getLogger

from .errors import MalformedMapProofError

# pylint: disable=C0103
logger = getLogger(__name__)


class OptionalEntry:
    """Optional entry is an entry of MapProof which can either miss a key or a key/value pair."""

    def __init__(self, key: Any, value: Optional[Any]):
        self.key = key
        self.value = value
        self.is_missing = not value  # False if the value is set.

    def __repr__(self) -> str:
        if self.is_missing:
            return "Missing [key: {}]".format(self.key)

        return "Entry [key: {}, value: {}]".format(self.key, self.value)

    @staticmethod
    def parse(data: Dict[str, Any]) -> "OptionalEntry":
        """Parsed an OptionalEntry from the provided JSON dict."""
        if data.get("missing"):
            return OptionalEntry(key=data["missing"], value=None)

        if data.get("key") and data.get("value"):
            return OptionalEntry(key=data["key"], value=data["value"])

        logger.warning("Failed to parse an OptionalEntry from the provided JSON dict.")
        raise MalformedMapProofError.malformed_entry(data)



"""
Exonum Python Light Client.

This library is designed to provide useful interfaces for Exonum blockchain.

The main modules you can be interested in:
  - client:
    Module with ExonumClient, the main entity to interact with Exonum. It will
    provide you with the methods to work with API, send transactions, etc.
  - message:
    Module with classes to generate transactions.
  - module_manager:
    This module contains ModuleManager entity to work with generated Protobuf classes
    (see [documentation](https://exonum.com/doc/version/latest/architecture/transactions/#serialization)
    if you are not familiar with the Exonum Protobuf workflow).
  - proofs:
    Module with ListProof and MapProof classes which can be used to verify proofs obtained from Exonum.
"""

from .client import ExonumClient
from .message import MessageGenerator, ExonumMessage
from .module_manager import ModuleManager
from .proofs import ListProof



""" Module with a MapProof Representation of the Branch Node. """
from logging import getLogger

from exonum_client_cis.crypto import Hash
from .constants import PROOF_PATH_SIZE
from .proof_path import ProofPath
from ..hasher import Hasher

# pylint: disable=C0103
logger = getLogger(__name__)


class BranchNode:
    """ MapProof representation of the branch node. """

    # Branch node contains 2 proof paths and 2 hashes:
    BRANCH_NODE_SIZE = 2 * (Hasher.HASH_SIZE + PROOF_PATH_SIZE)

    def __init__(self) -> None:
        self.raw = bytearray([0] * self.BRANCH_NODE_SIZE)

    @staticmethod
    def _verify_kind(kind: str) -> None:
        if kind not in ["left", "right"]:
            logger.warning("Incorrect child kind: %s. Should be one of these: 'left', 'right'.", kind)
            raise ValueError("Incorrect child kind: {}".format(kind))

    def _hash_slice(self, kind: str) -> slice:
        self._verify_kind(kind)
        start = 0 if kind == "left" else Hasher.HASH_SIZE

        return slice(start, start + Hasher.HASH_SIZE)

    def _path_slice(self, kind: str) -> slice:
        self._verify_kind(kind)
        start = 2 * Hasher.HASH_SIZE if kind == "left" else 2 * Hasher.HASH_SIZE + PROOF_PATH_SIZE

        return slice(start, start + PROOF_PATH_SIZE)

    def child_hash(self, kind: str) -> Hash:
        """Returns a stored child hash for the specified kind ("left" or "right")."""
        return Hash(bytes(self.raw[self._hash_slice(kind)]))

    def child_path(self, kind: str) -> ProofPath:
        """Returns a stored child path for the specified kind ("left" or "right")."""
        return ProofPath(self.raw[self._path_slice(kind)], 0)

    def set_child_path(self, kind: str, prefix: ProofPath) -> None:
        """Sets a child path for the specified kind ("left" or "right")."""
        self.raw[self._path_slice(kind)] = prefix.as_bytes()

    def set_child_hash(self, kind: str, child_hash: Hash) -> None:
        """Sets a child hash for the specified kind ("left" or "right")."""
        self.raw[self._hash_slice(kind)] = child_hash.value

    def set_child(self, kind: str, prefix: ProofPath, child_hash: Hash) -> None:
        """Sets a child (both path and hash) for specified kind ("left" or "right")."""
        self.set_child_path(kind, prefix)
        self.set_child_hash(kind, child_hash)

    def object_hash(self) -> Hash:
        """Returns a hash of the branch node."""
        data = bytearray()

        data += self.raw[self._hash_slice("left")]
        data += self.raw[self._hash_slice("right")]

        left_path_compressed = self.child_path("left").as_bytes_compressed()
        data += left_path_compressed

        right_path_compressed = self.child_path("right").as_bytes_compressed()
        data += right_path_compressed

        return Hasher.hash_map_branch(data)



"""Module with Hashing Utils for Proofs."""
from enum import IntEnum
import struct

# From pysodium import crypto_hash_sha256, crypto_hash_sha256_BYTES
from exonum_client_cis.crypto import HASH_BYTES_LEN, Hash

# Default hash value for empty ProofMapIndex.
EMPTY_MAP_HASH = bytes.fromhex("7324b5c72b51bb5d4c180f1109cfd347b60473882145841c39f3e584576296f9")


class Hasher:
    """Hasher class performs hashing operations on proof elements."""

    HASH_SIZE = HASH_BYTES_LEN
    DEFAULT_HASH = bytes([0] * HASH_BYTES_LEN)

    class HashTag(IntEnum):
        """Prefixes for different types of nodes."""

        BLOB = 0
        LIST_BRANCH_NODE = 1
        LIST_NODE = 2
        MAP_NODE = 3
        MAP_BRANCH_NODE = 4

    @staticmethod
    def hash_raw_data(data: bytes) -> Hash:
        """ SHA256 hash of the provided data. """

        return Hash.hash_data(data)

    @staticmethod
    def hash_node(left: Hash, right: Hash) -> Hash:
        """ Convenience method to obtain a hashed value of the merkle tree node. """

        data = struct.pack("<B", Hasher.HashTag.LIST_BRANCH_NODE) + left.value + right.value

        return Hash.hash_data(data)

    @staticmethod
    def hash_single_node(left: Hash) -> Hash:
        """ Convenience method to obtain a hashed value of the merkle tree node with one child. """

        data = struct.pack("<B", Hasher.HashTag.LIST_BRANCH_NODE) + left.value

        return Hash.hash_data(data)

    @staticmethod
    def hash_leaf(val: bytes) -> Hash:
        """ Convenience method to obtain a hashed value of the merkle tree leaf. """

        data = struct.pack("<B", Hasher.HashTag.BLOB) + val

        return Hash.hash_data(data)

    @staticmethod
    def hash_list_node(length: int, merkle_root: Hash) -> Hash:
        """
        Hash of the list object.
        ```text
        h = sha-256( HashTag::ListNode || len as u64 || merkle_root )
        ```
        """
        data = struct.pack("<BQ", Hasher.HashTag.LIST_NODE, length) + merkle_root.value

        return Hash.hash_data(data)

    @staticmethod
    def hash_map_node(root: Hash) -> Hash:
        """
        Hash of the map object.
        ```text
        h = sha-256( HashTag::MapNode || merkle_root )
        ```
        """
        data = struct.pack("<B", Hasher.HashTag.MAP_NODE) + root.value

        return Hash.hash_data(data)

    @staticmethod
    def hash_map_branch(branch_node: bytes) -> Hash:
        """
        Hash of the map branch node.
        ```text
        h = sha-256( HashTag::MapBranchNode || <left_key> || <right_key> || <left_hash> || <right_hash> )
        ```
        """
        data = struct.pack("<B", Hasher.HashTag.MAP_BRANCH_NODE) + branch_node

        return Hash.hash_data(data)

    @staticmethod
    def hash_single_entry_map(path: bytes, child_hash: Hash) -> Hash:
        """
        Hash of the map with single entry.
        ``` text
        h = sha-256( HashTag::MapBranchNode || <key> || <child_hash> )
        ```
        """
        data = struct.pack("<B", Hasher.HashTag.MAP_BRANCH_NODE) + path + child_hash.value

        return Hash.hash_data(data)



"""Proof Verification Module for Exonum `ProofListIndex`."""

from typing import Dict, List, Tuple, Any, Callable
import itertools
from logging import getLogger

from exonum_client_cis.crypto import Hash
from ..utils import is_field_hash, is_field_int, calculate_height
from ..hasher import Hasher
from .key import ProofListKey
from .errors import MalformedListProofError, ListProofVerificationError

# pylint: disable=C0103
logger = getLogger(__name__)


class HashedEntry:
    """ Element of a proof with a key and a hash. """

    def __init__(self, key: ProofListKey, entry_hash: Hash):
        self.key = key
        self.entry_hash = entry_hash

    @classmethod
    def parse(cls, data: Dict[Any, Any]) -> "HashedEntry":
        """ Creates a HashedEntry object from the provided dict. """
        if not isinstance(data, dict) or not is_field_hash(data, "hash"):
            err = MalformedListProofError.parse_error(str(data))
            logger.warning(
                "Could not parse `hash` from dict, which is required for HashedEntry object creation. %s", str(err)
            )
            raise err

        key = ProofListKey.parse(data)
        return HashedEntry(key, Hash(bytes.fromhex(data["hash"])))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HashedEntry):
            raise TypeError("Attempt to compare HashedEntry with an object of a different type.")
        return self.key == other.key and self.entry_hash == other.entry_hash


def _hash_layer(layer: List[HashedEntry], last_index: int) -> List[HashedEntry]:
    """ Takes a layer as a list of hashed entries and the last index as an int and returns a new layer. """
    new_len = (len(layer) + 1) // 2
    new_layer: List[HashedEntry] = []

    for i in range(new_len):
        left_idx = 2 * i
        right_idx = 2 * i + 1

        # Check if there are both right and left indices in the layer:
        if len(layer) > right_idx:
            # Verify that entries are in the correct order:
            if not layer[left_idx].key.is_left() or layer[right_idx].key.index != layer[left_idx].key.index + 1:
                err = MalformedListProofError.missing_hash()
                logger.warning(str(err))
                raise err

            left_hash = layer[left_idx].entry_hash
            right_hash = layer[right_idx].entry_hash
            new_entry = HashedEntry(layer[left_idx].key.parent(), Hasher.hash_node(left_hash, right_hash))
        else:
            # If there is an odd number of entries, the index of the last one should be equal to provided last_index:
            full_layer_length = last_index + 1
            if full_layer_length % 2 == 0 or layer[left_idx].key.index != last_index:
                err = MalformedListProofError.missing_hash()
                logger.warning(str(err))
                raise err

            left_hash = layer[left_idx].entry_hash
            new_entry = HashedEntry(layer[left_idx].key.parent(), Hasher.hash_single_node(left_hash))
        new_layer.append(new_entry)

    return new_layer


class ListProof:
    """ListProof class provides an interface to parse and verify proofs for ProofListIndex retrieved
    from the Exonum blockchain.

    Example workflow:

    >>> proof_json = {
    >>>     "proof": [
    >>>         {"index": 1, "height": 1, "hash": "eae60adeb5c681110eb5226a4ef95faa4f993c4a838d368b66f7c98501f2c8f9"}
    >>>     ],
    >>>     "entries": [[0, "6b70d869aeed2fe090e708485d9f4b4676ae6984206cf05efc136d663610e5c9"]],
    >>>     "length": 2,
    >>> }
    >>> expected_hash = "07df67b1a853551eb05470a03c9245483e5a3731b4b558e634908ff356b69857"
    >>> proof = ListProof.parse(proof_json)
    >>> result = proof.validate(bytes.fromhex(expected_hash))
    >>> assert result == [(0, stored_val)]
    """

    def __init__(
        self,
        proof: List[HashedEntry],
        entries: List[Tuple[int, Any]],
        length: int,
        value_to_bytes: Callable[[Any], bytes],
    ):
        """
        Constructor of the ListProof.
        It is not intended to be used directly, use ListProof.Parse instead.

        Parameters
        ----------
        proof : List[HashedEntry]
            Proof entries.
        entries: List[Tuple[int, Any]]
            Unhashed entries (leaves).
        length: int
            Length of the proof list.
        value_to_bytes: Callable[[str], bytes]
            A function that converts the stored value to bytes for hashing.
        """
        self._proof = proof
        self._entries = entries
        self._length = length
        self._value_to_bytes = value_to_bytes

    @classmethod
    def parse(cls, proof_dict: Dict[str, Any], value_to_bytes: Callable[[Any], bytes] = bytes.fromhex) -> "ListProof":
        """
        Method to parse ListProof from the dict.

        Expected dict format:

        >>>
        {
            'proof': [
                {'index': 1, 'height': 1, 'hash': 'eae60adeb5c681110eb5226a4ef95faa4f993c4a838d368b66f7c98501f2c8f9'}
            ],
            'entries': [
                [0, '6b70d869aeed2fe090e708485d9f4b4676ae6984206cf05efc136d663610e5c9']
            ],
            'length': 2
        }

        If no errors occured during parsing, a ListProof object will be returned.
        However, successfull parsing does not mean that the proof is not malformed (it only means that the provided
        dict structure matches the expected one).
        Actual checks for the proof contents correctness will be performed in the `validate` method.

        To convert value to bytes, ListProof attemts to use bytes.fromhex by default.
        If your type should be converted to bytes using Protobuf, you can generate a converter function with the use of
        `build_encoder_function` from encoder.py.
        Otherwise, you have to implement the converter function by yourself.

        Parameters
        ----------
        proof_dict : Dict[str, Any]
            Proof as a dict.
        value_to_bytes: Callable[[str], bytes]
            A function that converts the stored value to bytes for hashing.
            By default, `bytes.fromhex` is used.

        Raises
        ------
        MalformedListProofError
            If the structure of the provided dict does not match the expected one,
            an exception `MalformedListProofError` is raised.
        """
        if (
            not isinstance(proof_dict.get("proof"), list)
            or not isinstance(proof_dict.get("entries"), list)
            or not is_field_int(proof_dict, "length")
        ):
            err = MalformedListProofError.parse_error(str(proof_dict))
            logger.warning("The structure of the provided dict does not match the expected one. %s", str(err))
            raise err

        proof = [HashedEntry.parse(entry) for entry in proof_dict["proof"]]
        entries = [cls._parse_entry(entry) for entry in proof_dict["entries"]]
        length = proof_dict["length"]

        logger.debug("Successfully parsed ListProof from the dict.")
        return ListProof(proof, entries, length, value_to_bytes)

    def validate(self, expected_hash: Hash) -> List[Tuple[int, Any]]:
        """
        This method validates the provided proof against the given expected hash.

        Parameters
        ----------
        expected_hash: Hash
            Expected root hash.

        Returns
        -------
        result: List[Tuple[int, Any]]
            If the hash is correct, a list of the collected values with indices is returned.

        Raises
        ------
        ListProofVerificationError
            If verification fails, an exception `ListProofVerificationError` is raised.
        MalformedListProofError
            If the proof is malformed, an exception `MalformedListProofError` is raised.
        """
        if not isinstance(expected_hash, Hash):
            raise TypeError("`expected_hash` should be of type Hash.")

        tree_root = self._collect()

        calculated_hash = Hasher.hash_list_node(self._length, tree_root)
        if calculated_hash != expected_hash:
            logger.warning("Provided root hash does not match the calculated one.")
            raise ListProofVerificationError(expected_hash.value, calculated_hash.value)
        logger.debug("Successfully validated the provided proof against the given expected hash.")

        return self._entries

    @staticmethod
    def _parse_entry(data: List[Any]) -> Tuple[int, Any]:
        if not isinstance(data, list) or not len(data) == 2:
            err = MalformedListProofError.parse_error(str(data))
            logger.warning("Could not parse a list. %s", err)
            raise err
        return data[0], data[1]

    @staticmethod
    def _tree_height_by_length(length: int) -> int:
        if length == 0:
            return 0

        return calculate_height(length)

    @staticmethod
    def _check_duplicates(entries: List[Any]) -> None:
        for idx in range(1, len(entries)):
            if entries[idx][0] == entries[idx - 1][0]:
                err = MalformedListProofError.duplicate_key()
                logger.warning(str(err))
                raise err

    def _collect(self) -> Hash:
        def _hash_entry(entry: Tuple[int, Any]) -> HashedEntry:
            """ Creates a hash entry from the value. """
            key = ProofListKey(1, entry[0])
            entry_hash = Hasher.hash_leaf(self._value_to_bytes(entry[1]))
            return HashedEntry(key, entry_hash)

        def _split_hashes_by_height(
            hashes: List[HashedEntry], height: int
        ) -> Tuple[List[HashedEntry], List[HashedEntry]]:
            """ Splits a list of the hashed entries into two lists by the given height. """
            current = list(itertools.takewhile(lambda x: x.key.height == height, hashes))
            remaining = hashes[len(current) :]

            return current, remaining

        tree_height = self._tree_height_by_length(self._length)

        # Check an edge case when the list contains no elements:
        if tree_height == 0 and (not self._proof or not self._entries):
            err = MalformedListProofError.non_empty_proof()
            logger.warning(str(err))
            raise err

        # If there are no entries, the proof should contain only a single root hash:
        if not self._entries:
            if len(self._proof) != 1:
                if self._proof:
                    err = MalformedListProofError.missing_hash()
                    logger.warning(str(err))
                    raise err
                err = MalformedListProofError.unexpected_branch()
                logger.warning(str(err))
                raise err

            if self._proof[0].key == ProofListKey(tree_height, 0):
                return self._proof[0].entry_hash

            err = MalformedListProofError.unexpected_branch()
            logger.warning(str(err))
            raise err

        # Sort the entries and the proof:
        self._entries.sort(key=lambda el: el[0])
        self._proof.sort(key=lambda el: el.key)

        # Check that there are no duplicates:
        self._check_duplicates(self._entries)
        self._check_duplicates(self._proof)

        # Check that the hashes at each height have indices in the allowed range:
        for entry in self._proof:
            height = entry.key.height
            if height == 0:
                err = MalformedListProofError.unexpected_leaf()
                logger.warning(str(err))
                raise err

            # self._length -1 is the index of the last element at `height = 1`.
            # This index is divided by 2 with each new height:
            if height >= tree_height or entry.key.index > (self._length - 1) >> (height - 1):
                err = MalformedListProofError.unexpected_branch()
                logger.warning(str(err))
                raise err

        # Create the first layer:
        layer = list(map(_hash_entry, self._entries))
        hashes = list(self._proof)
        last_index = self._length - 1

        for height in range(1, tree_height):
            # Filter the hashes of the current height and the rest heights (to be processed later):
            hashes, remaining_hashes = _split_hashes_by_height(hashes, height)

            # Merge the current layer with the hashes that belong to this layer:
            layer = sorted(layer + hashes, key=lambda x: x.key)

            # Calculate a new layer:
            layer = _hash_layer(layer, last_index)

            # Size of the next layer is two times smaller:
            last_index //= 2

            # Make remaining_hashes hashes to be processed:
            hashes = remaining_hashes

        assert len(layer) == 1, "Result layer length is not 1"
        return layer[0].entry_hash



"""MapProof Module.

This module contains classes to work with MapProofs obtained from Exonum.

The most important ones are MapProofBuilder and CheckedMapProof"""
from .map_proof import MapProof, CheckedMapProof
from .map_proof_builder import MapProofBuilder
from .errors import MalformedMapProofError



"""ProofPath Module."""
from typing import Optional
from functools import total_ordering
from enum import IntEnum
from logging import getLogger

from ..utils import div_ceil, reset_bits, leb128_encode_unsigned

from .constants import KEY_SIZE, PROOF_PATH_SIZE
from .errors import MalformedMapProofError

# pylint: disable=C0103
logger = getLogger(__name__)


@total_ordering
class ProofPath:
    """ProofPath is a representation of the key in MapProof."""

    class _KeyPrefix(IntEnum):
        BRANCH = 0
        LEAF = 1
        VALUE = 2

    class _Positions(IntEnum):
        KIND_POS = 0
        KEY_POS = 1
        LEN_POS = KEY_SIZE + 1

    @staticmethod
    def parse(bits: str) -> "ProofPath":
        """
        This method parses a ProofPath from a string.

        Parameters
        -----------
        bits: str
            Sequence of '0' and '1' as a string.

        Returns
        -------
        ProofPath
            Parsed ProofPath object.

        Raises
        ------
        MalformedProofError
            If an input string is incorrect (too long, empty or contains unexpected symbols).
        """

        length = len(bits)
        if length == 0 or length > 8 * KEY_SIZE:
            error = "Incorrect MapProof path length: {}".format(length)
            logger.warning(error)
            raise MalformedMapProofError.malformed_entry(bits, error)

        data = [0] * KEY_SIZE

        for i, char in enumerate(bits):
            if char == "0":
                pass
            elif char == "1":
                data[i // 8] += 1 << (i % 8)
            else:
                error = "Unexpected MapProof path symbol: {}".format(char)
                logger.warning(error)
                raise MalformedMapProofError.malformed_entry(bits, error)

        data_bytes = bytes(data)

        proof_path = ProofPath.from_bytes(data_bytes)
        if length != 8 * KEY_SIZE:
            proof_path = proof_path.prefix(length)

        logger.debug("Successfully parsed a ProofPath from a string.")
        return proof_path

    @staticmethod
    def from_bytes(data_bytes: bytes) -> "ProofPath":
        """
        Builds ProofPath from a byte sequence.

        Parameters
        -----------
        data_bytes: bytes
            Array of bytes with ProofPath data.

        Returns
        -------
        ProofPath
            Parsed ProofPath object.

        Raises
        ------
        ValueError
            Length of provided array is not equal to KEY_SIZE constant.
        """
        if len(data_bytes) != KEY_SIZE:
            logger.warning("Wrong length of the provided byte sequence: expected %s, got %s", KEY_SIZE, len(data_bytes))
            raise ValueError("Incorrect data size")

        inner = bytearray([0] * PROOF_PATH_SIZE)

        inner[0] = ProofPath._KeyPrefix.LEAF
        inner[ProofPath._Positions.KEY_POS : ProofPath._Positions.KEY_POS + KEY_SIZE] = data_bytes[:]
        inner[ProofPath._Positions.LEN_POS] = 0

        return ProofPath(inner, 0)

    def __init__(self, data_bytes: bytearray, start: int):
        """ Constructor of ProofPath. Expects arguments to be cleaned already and does not check anything. """
        self.data_bytes = data_bytes
        self._start = start

    def __repr__(self) -> str:
        """ Conversion to a string. """
        bits_str = ""

        raw_key = self.raw_key()
        for byte_idx, chunk in enumerate(raw_key):
            # Range from 7 to 0 inclusively:
            for bit in range(7, -1, -1):
                i = byte_idx * 8 + bit
                if i < self.start() or i >= self.end():
                    bits_str += "_"
                else:
                    bits_str += "0" if (1 << bit) & chunk == 0 else "1"

            bits_str += "|"

        format_str = "ProofPath [ start: {}, end: {}, bits: {} ]".format(self.start(), self.end(), bits_str)
        return format_str

    def __len__(self) -> int:
        return self.end() - self.start()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ProofPath):
            raise TypeError("Attempt to compare ProofPath with an object of a different type.")
        return len(self) == len(other) and self.starts_with(other)

    def bit(self, idx: int) -> int:
        """Returns a bit of the path at the specified position."""
        pos = self.start() + idx
        chunk = self.raw_key()[(pos // 8)]
        bit = pos % 8
        return ((1 << bit) & chunk) >> bit

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, ProofPath):
            raise TypeError("Attempt to compare ProofPath with an object of a different type.")

        if self.start() != other.start():
            return NotImplemented

        if self.start() != 0:
            # The code below does not work if `self.start() % 8 != 0` without additional modifications:
            raise ValueError("Comparison is allowed only for paths with start =0")

        this_len = len(self)
        other_len = len(other)

        intersecting_bits = min(this_len, other_len)

        pos = self.common_prefix_len(other)

        if pos == intersecting_bits:
            return this_len < other_len

        return self.bit(pos) < other.bit(pos)

    def is_leaf(self) -> bool:
        """ Returns True if ProofPath is a leaf. Otherwise returns False """
        return self.data_bytes[0] == ProofPath._KeyPrefix.LEAF

    def start(self) -> int:
        """ Returns the index of the start bit. """
        return self._start

    def end(self) -> int:
        """ Returns the index of the end bit. """
        if self.is_leaf():
            return KEY_SIZE * 8

        return self.data_bytes[ProofPath._Positions.LEN_POS]

    def raw_key(self) -> bytes:
        """ Returns the stored key as raw bytes. """
        return bytes(self.data_bytes[ProofPath._Positions.KEY_POS : ProofPath._Positions.KEY_POS + KEY_SIZE])

    def set_end(self, end: Optional[int]) -> None:
        """ Sets the right border of the proof path. """
        if end is not None:
            self.data_bytes[0] = self._KeyPrefix.BRANCH
            self.data_bytes[self._Positions.LEN_POS] = end
        else:
            self.data_bytes[0] = self._KeyPrefix.LEAF
            self.data_bytes[self._Positions.LEN_POS] = 0

    def prefix(self, length: int) -> "ProofPath":
        """ Creates a copy of this path shortened to the specified length. """

        end = self._start + length
        key_len = KEY_SIZE * 8

        if end >= key_len:
            err_msg = f"Length of the prefix ({end}) should not be greater than KEY_SIZE * 8 ({key_len})."
            logger.warning(err_msg)
            raise ValueError(err_msg)

        key = ProofPath(bytearray(self.data_bytes), self._start)
        key.set_end(end)

        return key

    def match_len(self, other: "ProofPath", from_bit: int) -> int:
        """ Returns the length of the common segment. """
        if self.start() != other.start():
            logger.warning("Misaligned bit ranges: %s != %s", self.start(), other.start())
            raise ValueError("Misaligned bit ranges")

        if from_bit < self.start() or from_bit > self.end():
            err_msg = f"Incorrect from_bit value: {from_bit}"
            logger.warning(err_msg)
            raise ValueError(err_msg)

        len_to_the_end = min(len(self), len(other))
        for i in range(from_bit, len_to_the_end):
            if self.bit(i) != other.bit(i):
                return i

        return len_to_the_end

    def common_prefix_len(self, other: "ProofPath") -> int:
        """ Returns the length of the common prefix. """
        if self.start() == other.start():
            return self.match_len(other, self.start())

        return 0

    def starts_with(self, other: "ProofPath") -> bool:
        """ Returns True if `other` is a prefix of `self`. Otherwise returns False. """
        return self.common_prefix_len(other) == len(other)

    def as_bytes(self) -> bytes:
        """ Represents a path as bytes according to the Merkledb implementation. """

        return bytes(self.data_bytes)

    def as_bytes_compressed(self) -> bytes:
        """ Represents a path as compressed bytes using les128 algorigthm. """
        bits_len = self.end()
        whole_bytes_len = div_ceil(bits_len, 8)

        key = self.raw_key()[0:whole_bytes_len]

        result = bytearray()
        result += leb128_encode_unsigned(bits_len)
        result += key

        # Trim insignificant bits in the last byte:
        bits_in_last_byte = bits_len % 8
        if whole_bytes_len > 0 and bits_in_last_byte != 0:
            tail = self.end() % 8
            result[-1] = reset_bits(result[-1], tail)

        return bytes(result)



"""
Exonum Client Module.

This module provides you with two handy classes:
  - ExonumClient: The main entity to interact with the Exonum blockchain.
  - Subscriber: An entity that can be used to receive signals on creation of a new block.
"""
from typing import Optional, Any, Callable, Union, Dict, Tuple

import json
from logging import getLogger
from threading import Thread
from urllib.parse import urlencode
from websocket import WebSocket

from .api import ServiceApi, PublicApi, PrivateApi, ProtobufApi
from .protobuf_loader import ProtobufLoader
from .message import ExonumMessage

# pylint: disable=C0103
logger = getLogger(__name__)


class Subscriber:
    """ Subscriber objects are used to subscribe to Exonum blocks via websockets. """

    # constants
    SUBSCRIPTION_WEBSOCKET_URI = "ws://{}:{}/api/explorer/v1/{}/subscribe"
    SENDING_WEBSOCKET_URI = "ws://{}:{}/api/explorer/v1/ws"
    SUBSCRIPTION_TYPES = ["blocks", "transactions"]

    # Type of the received data (it can be either bytes or a string).
    ReceiveType = Union[bytes, str]
    # Type of the `Callback` (`Callable` that takes `ReceiveType` as an argument and produces nothing).
    CallbackType = Callable[[ReceiveType], None]

    def __init__(
        self, address: str, port: int, subscription_type: Optional[str] = None, filters: Optional[Dict[str, Any]] = None
    ):
        """Subscriber constructor.

        Parameters
        ----------
        address: str
            IP address of the Exonum node.
        port: int
            Port of the exonum node.
        subscription_type: Optional[str]
            Type of subscription: "blocks" or "transactions". If not given,
            it is assumed that subscriber is used to send transactions.
        filters: Optional[Dict[str, Any]]
            Dictionary of filters, such as 'service_id' and 'tx_id' for transactions.
        """
        if not subscription_type:
            self._address = self.SENDING_WEBSOCKET_URI.format(address, port)
        else:
            if subscription_type not in Subscriber.SUBSCRIPTION_TYPES:
                err = ValueError(
                    f"Subscription type must be one of these: {self.SUBSCRIPTION_TYPES}, "
                    f"while {subscription_type} is given."
                )
                logger.error("Error occurred during subscriber initialization: %s", err)
                raise err
            parameters = "?" + urlencode(filters) if filters else ""
            self._address = self.SUBSCRIPTION_WEBSOCKET_URI.format(address, port, subscription_type) + parameters
        self._is_running = False
        self._connected = False
        self._ws_client = WebSocket()
        self._thread = Thread(target=self._event_processing)
        self._handler: Optional[Subscriber.CallbackType] = None

    def __enter__(self) -> "Subscriber":
        self.connect()

        return self

    def __exit__(self, exc_type: Optional[type], exc_value: Optional[Any], exc_traceback: Optional[object]) -> None:
        self.stop()

    def connect(self) -> None:
        """Connects the subscriber to the Exonum, so it will be able to receive events. """
        self._ws_client.connect(self._address)
        self._connected = True

    def set_handler(self, handler: "Subscriber.CallbackType") -> None:
        """Sets the handler """
        self._handler = handler

    def run(self) -> None:
        """Runs the subscriber thread. It will call the provided handler on every new block. """
        try:
            self._is_running = True
            self._thread.setDaemon(True)
            self._thread.start()
            logger.debug("Subscriber thread started successfully.")
        except RuntimeError as error:
            logger.error("Error occurred during running subscriber thread: %s", error)

    def _event_processing(self) -> None:
        while self._is_running:
            data = self._ws_client.recv()
            if data and self._handler:
                self._handler(data)

    def wait_for_new_event(self) -> None:
        """ Waits until a new event (block or transaction) is ready. Please note that this method is a blocking one. """
        if self._is_running:
            print("Subscriber is already running...")
        else:
            self._ws_client.recv()

    def stop(self) -> None:
        """Closes connection with the websocket and, if the thread is running, joins it. """
        if self._is_running:
            self._is_running = False

        if self._connected:
            self._ws_client.close()
            self._connected = False

        if self._thread.is_alive():
            self._thread.join()

    def send_transaction(self, message: ExonumMessage) -> str:
        """
        Sends a transaction into an Exonum node via WebSocket.
        Example:
        >>> response = client.send_websocket_transaction(message)
        >>> print(response)
        {"result":"success","response":{"tx_hash":"48b7d71d388f3c2dfa665bcf837e1fa0417ca559fb0163533ea72de6319e61ca"}}
        Parameters
        ----------
        message: ExonumMessage
            Prepared and signed an Exonum message.
        Returns
        -------
        result: str
            Result of the WebSocket request.
            If a transaction is correct and it is accepted, it will contain a JSON with a hash of the transaction.
        """
        body_raw = message.signed_raw()
        if body_raw is None:
            logger.critical("Attempt to send an unsigned message through websocket.")
            raise RuntimeError("Attempt to send an unsigned message.")
        data = json.dumps({"type": "transaction", "payload": {"tx_body": body_raw.hex()}})

        ws_client = WebSocket()
        ws_client.connect(self._address)
        ws_client.send(data)
        response = ws_client.recv()
        ws_client.close()

        return response


class ExonumClient:
    """ExonumClient class is capable of interaction with ExonumBlockchain.

    All the methods that perform requests to the Exonum REST API return a requests.Response object.
    So a user should manually verify that the status code of the request is correct and get the contents
    of the request via `response.json()`.

    Since ExonumClient uses `requests` library for the communication, user should expect that every
    method that performs an API call can raise a `requests` exception (e.g. `requests.exceptions.ConnectionError`).

    Example usage:

    >>> client = ExonumClient(hostname="127.0.0.1", public_api_port=8080, private_api_port=8081)
    >>> health_info = client.public_api.health_info().json()
    {'consensus_status': 'Enabled', 'connected_peers': 0}
    >>> user_agent = client.public_api.user_agent().json()
    exonum 0.13.0-rc.2/rustc 1.37.0 (eae3437df 2019-08-13)
    """

    def __init__(self, hostname: str, public_api_port: int = 80, private_api_port: int = 81, ssl: bool = False):
        """
        Constructor of ExonumClient.

        Parameters
        ----------
        hostname: str
            Examples: '127.0.0.1', 'www.some_node.com'.
        public_api_port: int
            Public API port of an Exonum node.
        private_api_port: int
            Private API port of an Exonum node.
        ssl: bool
            If True, HTTPS protocol is used for communication, otherwise HTTP is used.
        """
        self.schema = "https" if ssl else "http"
        self.hostname = hostname
        self.public_api_port = public_api_port
        self.private_api_port = private_api_port

        self.public_api = PublicApi(hostname, public_api_port, self.schema)
        self.private_api = PrivateApi(hostname, private_api_port, self.schema)
        self.protobuf_api = ProtobufApi(hostname, public_api_port, self.schema)

    def __repr__(self) -> str:
        """ Conversion to a string. """
        d = {
            "object": f"<{self.__class__.__name__} instance at {id(self)}>",
            "host": self.hostname,
            "public_port": str(self.public_api_port),
            "private_port": str(self.private_api_port),
        }

        return json.dumps(d, indent=2)

    def service_private_api(self, service_name: str) -> ServiceApi:
        """Creates an instances of ServiceApi to interact with private API of a service.

        Parameters
        ----------
        service_name: str
            Name of a service.

        Returns
        -------
        service_api: ServiceApi
            An instance of ServiceApi for private API.
        """
        return ServiceApi(service_name, self.hostname, self.private_api_port, self.schema)

    def service_public_api(self, service_name: str) -> ServiceApi:
        """Creates an instances of ServiceApi to interact with public API of a service.

        Parameters
        ----------
        service_name: str
            Name of a service.

        Returns
        -------
        service_api: ServiceApi
            An instance of ServiceApi for public API.
        """
        return ServiceApi(service_name, self.hostname, self.public_api_port, self.schema)

    def service_apis(self, service_name: str) -> Tuple[ServiceApi, ServiceApi]:
        """Creates a tuple of ServiceApi instances to interact with public and private API of a service.

        Parameters
        ----------
        service_name: str
            Name of a service.

        Returns
        -------
        service_apis: (ServiceApi, ServiceApi)
            Tuple of ServiceApi instances: first for public API and second for private.
        """
        return self.service_public_api(service_name), self.service_private_api(service_name)

    def protobuf_loader(self) -> ProtobufLoader:
        """
        Creates a ProtobufLoader from the current ExonumClient object.

        See ProtobufLoader docs for more details.

        Example:

        >>> with client.protobuf_loader("blocks") as loader:
        >>>     loader.load_main_proto_files()
        >>>     loader.load_service_proto_files(0, "exonum-supervisor:0.13.0-rc.2")
        """
        return ProtobufLoader(self.protobuf_api)

    def create_subscriber(self, subscription_type: str) -> Subscriber:
        """
        Creates a Subscriber object from the current ExonumClient object.

        See Subscriber docs for details.

        Example:

        >>> with client.create_subscriber("blocks") as subscriber:
        >>>     subscriber.wait_for_new_event()

        Parameters
        ----------
        subscription_type: str
            Sets type of subscription: "blocks" or "transactions".
        """
        subscriber = Subscriber(self.hostname, self.public_api_port, subscription_type)
        return subscriber



"""This module is capable of creating and signing Exonum transactions."""

from typing import Dict, Optional, Tuple, Any
import json
from logging import getLogger

from google.protobuf.message import Message as ProtobufMessage, DecodeError as ProtobufDecodeError

from .crypto import PublicKey, Hash, Signature, KeyPair, SignStream
from .module_manager import ModuleManager

# pylint: disable=C0103
logger = getLogger(__name__)


class MessageGenerator:
    """MessageGenerator is a class which helps you create transactions.
    It is capable of transforming a Protobuf message object into an Exonum transaction
    with a set of the required metadata.

    Example of usage:
    >>> instance_id = ... # Get the ID of the desired service instance.
    >>> artifact_name = ... # Get the name of the artifact (not the instance).
    >>> cryptocurrency_message_generator = MessageGenerator(instance_id, artifact_name) # Create a message generator.
    >>> create_wallet_alice = cryptocurrency_module.CreateWallet() # Create a Protobuf message.
    >>> create_wallet_alice.name = "Alice1" # Fill the Protobuf message manually.

    Then you can transform the Protobuf message into an Exonum transaction.

    >>> create_wallet_alice_tx = cryptocurrency_message_generator.create_message(create_wallet_alice)
    >>> create_wallet_alice_tx.sign(0, keypair) # You should sign the message before sending.
    >>> client.public_api.send_transaction(create_wallet_alice_tx)
    """

    def __init__(self, instance_id: int, artifact_name: str):
        """MessageGenerator constructor.

        Parameters
        ----------
        instance_id: int
            ID of the desired Exonum service instance.
        artifact_name: str
            The name of the service artifact you want to communicate with.
            The name should be in the format provided by Exonum, like 'exonum-cryptocurrency-advanced:0.13.0-rc.2'.
        """
        self._instance_id = instance_id

        self._artifact_name = artifact_name
        self._message_ids: Dict[str, int] = dict()

        service_module = ModuleManager.import_service_module(artifact_name, "service")
        for i, message in enumerate(service_module.DESCRIPTOR.message_types_by_name):
            self._message_ids[message] = i

    def create_message(self, message: ProtobufMessage) -> "ExonumMessage":
        """Method to convert a Protobuf message into an Exonum message.

        Parameters
        ----------
        message: google.protobuf.message.Message
            A Protobuf message.

        Returns
        -------
        exonum_message: ExonumMessage
            Exonum message object.
        """

        tx_name = type(message).__name__
        return ExonumMessage(self._instance_id, self._message_ids[tx_name], message)


class ExonumMessage:
    """Generic Exonum transaction class.

    Exonum message can be created:
     - by using MessageGenerator (if you want to send a transaction to the Exonum blockchain)
     - by using ExonumMessage.from_hex (if you want to parse a retrieved transaction).

    Example workflow:

    Sending a message:

    >>> instance_id = ... # Get the ID of the desired service instance.
    >>> artifact_name = ... # Get the name of the artifact (not the instance).
    >>> cryptocurrency_message_generator = MessageGenerator(instance_id, artifact_name) # Create a message generator.
    >>> create_wallet_alice = cryptocurrency_module.CreateWallet() # Create a Protobuf message.
    >>> create_wallet_alice.name = "Alice1" # Fill the Protobuf message manually.
    >>> create_wallet_alice_tx = cryptocurrency_message_generator.create_message(create_wallet_alice)
    >>> create_wallet_alice_tx.sign(0, keypair) # You should sign the message before sending.
    >>> client.public_api.send_transaction(create_wallet_alice_tx)

    Parsing a message:

    >>> message_hex = ... # Retrieve the message as a hexadecimal string.
    >>> artifact_name = ... # Get the name of the artifact (not the instance).
    >>> transaction_name = "CreateWallet" # Get the name of the transaction.
    >>> parsed_message = ExonumMessage.from_hex(message_hex, artifact_name, transaction_name)
    >>> assert parsed_message.validate() # Verify the signature of the retrieved message.

    Other methods:
    >>> message = ExonumMessage.from_hex(...)
    >>> author = message.author() # Get the author's public key.
    >>> tx_hash = message.hash() # Get the transaction hash.
    >>> signature = message.signature() # Get the transaction signature.
    >>> any_tx_raw = message.any_tx_raw() # Get AnyTx of the message serialized to bytes.
    >>> signed_tx_raw = message.signed_tx_raw() # Get SignedMessage of the message serialized to bytes.
    >>> tx_json = message.pack_into_json() # Create a JSON with the transaction in the format expected by Exonum.
    """

    def __init__(self, instance_id: int, message_id: int, msg: ProtobufMessage, prebuilt: Optional[bytes] = None):
        """Exonum message constructor. It is not intended to be used directly, see `MessageGenerator.create_message`
        and `ExonumMessage.from_hex` instead."""
        # Author's public key as bytes.
        self._author: Optional[PublicKey] = None
        # ID of the service instance.
        self._instance_id = instance_id
        # ID of the message (to be set in the CallInfo).
        self._message_id = message_id
        # Protobuf message.
        self._msg = msg
        # Signature as bytes.
        self._signature: Optional[Signature] = None
        # AnyTx message serialized to bytes.
        # SignedMessage serialized to bytes.
        self._signed_tx_raw: Optional[bytes] = None

        # If we parse the received message, we do not have to build anything.
        if prebuilt is None:
            self._any_tx_raw = self._build_message()
        else:
            self._any_tx_raw = prebuilt

    @classmethod
    def from_hex(
        cls, message_hex: str, artifact_name: str, tx_name: str
    ) -> Optional["ExonumMessage"]:
        """Attempts to parse an Exonum message from a serialized hexadecimal string.

        Parameters
        ----------
        message_hex: str
            Serialized message as a hexadecimal string.
        artifact_name: str
            The name of the service artifact you want to communicate with,
            e.g. 'exonum-cryptocurrency-advanced:0.13.0-rc.2'.
        tx_name: str
            The name of the transaction to be parsed, e.g. 'CreateWallet'.

        Returns
        -------
        parsed_message: Optional[ExonumMessage]
            If parsing is successfull, an ExonumMessage object is returned.
            Otherwise the returned value is None.
        """
        try:
            signed_msg, exonum_msg, decoded_msg = cls._deserialize_message(
                message_hex, artifact_name, tx_name
            )

            service_id = exonum_msg.any_tx.call_info.instance_id
            message_id = exonum_msg.any_tx.call_info.method_id
            signature = signed_msg.signature.data[:]
            author = signed_msg.author.data[:]

            exonum_message = cls(service_id, message_id, decoded_msg, prebuilt=exonum_msg.any_tx.SerializeToString())

            cls._set_signature_data(exonum_message, author, signature, bytes.fromhex(message_hex))
            logger.debug(
                "Exonum message (ID: %s) from the service artifact '%s' " "for transaction '%s' parsed successfully.",
                message_id,
                artifact_name,
                tx_name,
            )
            return exonum_message
        except ProtobufDecodeError as e:
            logger.error(
                "Failed to parse an Exonum message from the service artifact '%s' " "for transaction '%s'. Error: %s.",
                artifact_name,
                tx_name,
                str(e),
            )
            return None

    def sign(self, chain_id: int, keys: KeyPair) -> None:
        """Signs the message with the provided pair of keys.

        Please note that signing is required before sending a message to the Exonum blockchain.

        Parameters
        ----------
        keys: exonum.crypto.KeyPair
            A pair of public_key and secret_key as bytes.
        keys: int
            A chain identifier.
        """

        public_key, secret_key = keys.public_key, keys.secret_key
        self._author = public_key

        consensus_mod = ModuleManager.import_main_module("consensus")
        types_mod = ModuleManager.import_main_module("types")

        signed_message = consensus_mod.SignedMessage()
        signed_message.payload = self._any_tx_raw
        signed_message.author.CopyFrom(types_mod.PublicKey(data=public_key.value))
        signed_message.chain_id_bytes = chain_id.to_bytes(8, byteorder='little')

        data_to_sign = signed_message.chain_id_bytes + signed_message.payload
        signature = Signature.sign(data_to_sign, secret_key)

        signed_message.signature.CopyFrom(types_mod.Signature(data=signature.value))

        self._signature = signature

        self._signed_tx_raw = bytes(signed_message.SerializeToString())

        logger.debug(
            "Successfully signed the message (message ID: %s, service instance ID: %s): public_key='%s'.",
            self._message_id,
            self._instance_id,
            public_key,
        )

    def validate(self) -> bool:
        """
        Validates the message.
        Checks if the transaction signature is correct.
        :return: bool
        """
        if self._signature is None or self._author is None:
            return False

        try:
            consensus_mod = ModuleManager.import_main_module("consensus")

            signed_msg = consensus_mod.SignedMessage()
            signed_msg.ParseFromString(self._signed_tx_raw)

            return self._signature.verify(signed_msg.chain_id_bytes + signed_msg.payload, self._author)
        except (ProtobufDecodeError, ValueError) as e:
            logger.error(
                "Failed to parse Exonum message (message ID: %s, service instance ID: %s): public_key='%s'. Error: %s",
                self._message_id,
                self._instance_id,
                self._author,
                str(e),
            )
            return False

    def pack_into_json(self) -> str:
        """Packs a serialized signed message into the JSON format expected by Exonum.

        Please note that this method does not serialize the message to JSON.

        Returns
        -------
        json_message: str
            String with a JSON representation of the serialized message.

        Raises
        ------
        RuntimeError
            An error will be raised on attempt to call `pack_into_json` with an unsigned message.
        """
        if self._signed_tx_raw is None:
            logger.critical("Attempt to call `to_json` on an unsigned message into JSON format.")
            raise RuntimeError("Attempt to pack an unsigned message.")
        return json.dumps({"tx_body": self._signed_tx_raw.hex()}, indent=4)

    def hash(self) -> Hash:
        """Returns a hash of the message. If the message is not signed, a hash of an empty message will be returned."""
        return Hash.hash_data(self._signed_tx_raw)

    # Getters section.

    def author(self) -> Optional[PublicKey]:
        """Returns an author's public key. If the author is not set, returns None."""
        return self._author

    def signature(self) -> Optional[Signature]:
        """Returns a signature. If the message is not signed, returns None."""
        return self._signature

    def any_tx_raw(self) -> bytes:
        """Returns a serialized AnyTx message as bytes."""
        return self._any_tx_raw

    def signed_raw(self) -> Optional[bytes]:
        """Returns a serialized SignedMessage as bytes. If the message is not signed, returns None."""
        return self._signed_tx_raw

    def _set_signature_data(self, author: bytes, signature: bytes, raw: bytes) -> None:
        self._author = PublicKey(author)
        self._signature = Signature(signature)
        self._signed_tx_raw = raw

    def _build_message(self) -> bytes:
        """Builds a raw AnyTx message."""
        runtime_mod = ModuleManager.import_main_module("runtime")
        consensus_mod = ModuleManager.import_main_module("consensus")

        serialized_msg = self._msg.SerializeToString()

        call_info = runtime_mod.CallInfo()
        call_info.instance_id = self._instance_id
        call_info.method_id = self._message_id

        any_tx = runtime_mod.AnyTx()
        any_tx.call_info.CopyFrom(call_info)
        any_tx.arguments = serialized_msg

        exonum_message = consensus_mod.ExonumMessage()
        exonum_message.any_tx.CopyFrom(any_tx)

        return exonum_message.SerializeToString()

    @staticmethod
    def _deserialize_message(
        message_hex: str, artifact_name: str, tx_name: str
    ) -> Tuple[Any, Any, Any]:
        """Takes a serialized message as an argument and returns a tuple
        [SignedMessage, ExonumMessage,DecodedMessage]."""

        # Load modules and prepare an expected message class for parsing.
        consensus_mod = ModuleManager.import_main_module("consensus")
        service_mod = ModuleManager.import_service_module(artifact_name, "service")
        transaction_class = getattr(service_mod, tx_name)

        # Convert a message from hex to bytes.
        tx_raw = bytes.fromhex(message_hex)

        # Parse SignedMessage.
        signed_msg = consensus_mod.SignedMessage()
        signed_msg.ParseFromString(tx_raw)

        # Parse ExonumMessage from the SignedMessage's payload.
        exonum_msg = consensus_mod.ExonumMessage()
        exonum_msg.ParseFromString(signed_msg.payload)

        # Parse an expected message from ExonumMessage's AnyTx arguments.
        decoded_msg = transaction_class()
        decoded_msg.ParseFromString(exonum_msg.any_tx.arguments)

        return signed_msg, exonum_msg, decoded_msg



"""Common Utils for Proofs Modules."""
from typing import Any, Dict, Callable, Optional
from logging import getLogger
import re

# Those utils are internal and very simple, so they don't require docs.
# pylint: disable=missing-docstring
# pylint: disable=C0103
logger = getLogger(__name__)


def is_dict(json: Any) -> bool:
    return isinstance(json, dict)


def is_field_dict(json: Dict[Any, Any], field: str) -> bool:
    return isinstance(json.get(field), dict)


def is_field_hash(json: Dict[Any, Any], field: str) -> bool:
    field_value = json.get(field)
    return isinstance(field_value, str) and bool(re.match(r"^[0-9A-Fa-f]{64}$", field_value))


def is_field_hash_or_none(json: Dict[Any, Any], field: str) -> bool:
    return not json.get(field) or is_field_hash(json, field)


def is_field_int(json: Dict[Any, Any], field: str) -> bool:
    return isinstance(json.get(field), int)


def is_field_convertible(json: Dict[Any, Any], field: str, value_to_bytes: Callable[[Any], bytes]) -> bool:
    try:
        if not json.get(field):
            return False
        value_to_bytes(json[field])
        return True
    except ValueError:
        logger.warning("Field '%s' is not convertible.", field)
        return False


def to_bytes(hex_data: str) -> Optional[bytes]:
    if not hex_data:
        return None

    return bytes.fromhex(hex_data)


def calculate_height(number: int) -> int:
    if number < 0:
        logger.warning("Number %s is used for tree height calculation and cannot be less than zero.", number)
        raise ValueError(f"Number {number} is less than zero.")
    if number == 0:
        return 1

    # Height of the tree is the amount of bits in the number that represents
    # the next power of two for the given number.
    # The next power of two is calculated as 1 << (number - 1).bit_length().
    # Thus, (number - 1).bit_length() + 1 is the required amount of bits.
    trailing_zeroes_amount = (number - 1).bit_length()

    return trailing_zeroes_amount + 1


def div_ceil(dividend: int, divider: int) -> int:
    return (dividend + divider - 1) // divider


def reset_bits(value: int, pos: int) -> int:
    """ Resets bits higher than the given pos. """
    reset_bits_mask = ~(255 << pos)
    value &= reset_bits_mask
    return value


def leb128_encode_unsigned(value: int) -> bytes:
    """ Encodes an unsigned number with leb128 algorithm. """
    if value < 0:
        logger.warning("Value passed to LEB128 for unsigned integers should be non-negative.")
        raise ValueError("Value should be non-negative")

    result = []
    while True:
        # Lower 7 bits of value:
        byte = value & 0x7F
        value >>= 7

        if value != 0:  # More bytes to come.
            byte |= 0x80  # Set high order bit.

        result.append(byte)

        if value == 0:
            break

    return bytes(result)



""" ProofListKey Module. """
from typing import Dict, Any
from functools import total_ordering
from logging import getLogger

from ..utils import is_field_int
from .errors import MalformedListProofError

# pylint: disable=C0103
logger = getLogger(__name__)


@total_ordering
class ProofListKey:
    """ A structure that represents a key in the list proof. """

    def __init__(self, height: int, index: int):
        self.height = height
        self.index = index

    @classmethod
    def parse(cls, data: Dict[Any, Any]) -> "ProofListKey":
        """ Parses ProofListKey from dict. """
        if not is_field_int(data, "index") or not is_field_int(data, "height"):
            err = MalformedListProofError.parse_error(str(data))
            logger.warning(str(err))
            raise err

        return cls(data["height"], data["index"])

    def is_left(self) -> bool:
        """ Returns true if this key is in the left branch of the proof. """
        return self.index % 2 == 0

    def parent(self) -> "ProofListKey":
        """ Returns the parent key for this key. """
        return ProofListKey(self.height + 1, self.index >> 1)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ProofListKey):
            raise TypeError("Attempt to compare ProofListKey with an object of a different type.")
        return self.index == other.index and self.height == other.height

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, ProofListKey):
            raise TypeError("Attempt to compare ProofListKey with an object of a different type.")

        # Try to compare by height, otherwise compare by index.
        if self.height != other.height:
            return self.height < other.height

        return self.index < other.index



"""Exonum Proofs Module.

This subpackage provides interfaces for parsing and verifying proofs obtained from Exonum.

For list proofs use the ListProof class, for map proofs use the MapProof and MapProofBuilder classes.

Since keys and values are hashed during the proof verification, and the algorigthm of serialization to bytes
may vary for different indices, you will have to provide encoder functions to the Proof object.

If key/value is serialized to bytes using Protobuf, function 'build_encoder_function' will generate
an encoder function for you."""
from .list_proof import *
from .map_proof import *
from .encoder import build_encoder_function



#!/usr/bin/env python
"""Setup Script for the Exonum CIS Python Light Client."""
import setuptools

INSTALL_REQUIRES = ["protobuf", "pysodium", "requests", "websocket-client-py3"]

PYTHON_REQUIRES = ">=3.4"

with open("README.md", "r") as readme:
    LONG_DESCRIPTION = readme.read()

setuptools.setup(
    name="exonum-client-cis",
    version="0.4.0.dev6",
    author="The Exonum team",
    author_email="contact@exonum.com",
    description="Exonum CIS Python Light Client",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=[
        "exonum_client_cis",
        "exonum_client_cis.proofs",
        "exonum_client_cis.proofs.map_proof",
        "exonum_client_cis.proofs.list_proof",
    ],
    install_requires=INSTALL_REQUIRES,
    python_requires=PYTHON_REQUIRES,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Security :: Cryptography",
    ],
)



"""Common Constants for the MapProof Module."""
from ..hasher import Hasher

# Size in bytes of Hash. Equal to the hash function output (32):
KEY_SIZE = Hasher.HASH_SIZE
# Size in bytes of ProofPath:
PROOF_PATH_SIZE = KEY_SIZE + 2
